import os
#os.environ["CUDA_LAUNCH_BLOCKING"] = "1"
os.environ["WANDB_MODE"] = "disabled"
os.environ["WANDB_DISABLED"] = "true"
from transformers import AutoTokenizer, AutoModelForMaskedLM, AutoModelForSequenceClassification,pipeline,\
DataCollatorForLanguageModeling, DataCollatorWithPadding, Trainer, TrainingArguments, AutoModel, EvalPrediction, AutoConfig
from datasets import load_dataset, Dataset, Features, load_metric, DatasetDict
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import numpy as np
from typing import Union
from transformers import TrainerCallback
import pandas as pd

 
from datasets import Dataset, DatasetDict
from datasets import load_dataset
from datasets import load_dataset, Dataset
import pandas as pd
from datasets import load_dataset, Dataset
from sklearn.model_selection import train_test_split

class DataLoader:
    def __init__(self, label2_id):
        self.label2_id = label2_id # {'entailment': 0, 'neutral': 1, 'contradiction': 2}
    
    def _print_dataset_status(self, dataset, task_type):
        def get_sample_count(dataset, task_type):
            if task_type == 'nli':
                return len(dataset['premise'])
            else:
                return len(dataset)

        print(f"Sampled {get_sample_count(dataset['train'], task_type)} training samples!")
        # Print status for validation set (if applicable)
        if "validation" in dataset:
            print(f"Sampled {get_sample_count(dataset['validation'], task_type)} validation samples!")

    def _convert_labels_to_numeric(self, dataset, task_type):
        """Converts non-numeric labels to numeric labels using `self.label2_id`."""
        if task_type == 'nli' and self.label2_id:
            def convert_labels(sentence):
                sentence['label'] = self.label2_id[sentence['label']] if sentence['label'] in self.label2_id else sentence['label']
                return sentence
            dataset = dataset.map(convert_labels)
        return dataset


    def _load_csv_data(self, dataset_path, task_type, num_percentage_validation, val_dataset):
        # Prepare data files for loading
        data_files = {'train': dataset_path}
        if val_dataset:
            data_files['validation'] = val_dataset

        if task_type=="mlm":
            dataset = load_dataset('csv', data_files=data_files, encoding="utf-8-sig")
        # Load dataset with appropriate columns
        elif task_type=='nli':
            dataset = load_dataset('csv', data_files=data_files, encoding="utf-8-sig")
        if num_percentage_validation and not val_dataset:
            if not (0 < num_percentage_validation < 1):
                raise ValueError("num_percentage_validation must be between 0 and 1")
            

            dataset = dataset['train'].train_test_split(test_size=num_percentage_validation)
            dataset['validation'] = dataset['test']
            del dataset['test']
                
        if "validation" in dataset:
            dataset['validation'] = self._convert_labels_to_numeric(dataset['validation'], task_type)
        
        # Convert labels to numeric if needed
        dataset['train'] = self._convert_labels_to_numeric(dataset['train'], task_type)

        return dataset


    def _prepare_dict_dataset(self, data_dict, task_type, num_percentage_validation, val_dataset):
        """
        Load a dataset from a dictionary for NLI or MLM, with optional validation set splitting.
        
        Args:
            data_dict (dict): A dictionary containing the data. For NLI, it should have 'premise', 'hypothesis', and 'label'.
                              For MLM, it should have 'text'.
            task_type (str): Either 'nli' for Natural Language Inference or 'mlm' for Masked Language Modeling.
            num_percentage_validation (float): The percentage of data to use for validation if no validation set is provided.
            validate (bool): Whether to create a validation set if it doesn't exist in the data.
            val_dataset (dict): A validation dataset, if it exists separately.
        
        Returns:
            DatasetDict: A DatasetDict object containing 'train' and optionally 'validation' splits.
        """

        if 'train' not in data_dict:
            raise ValueError("The dictionary must contain a 'train' key with the data.")

        if task_type == 'nli':
            # Ensure the NLI format has 'premise', 'hypothesis', and 'label' keys
            if not all(key in data_dict['train'] for key in ['premise', 'hypothesis', 'label']):
                raise ValueError("For NLI, the dictionary must contain 'premise', 'hypothesis', and 'label' keys.")

            if "validation" in data_dict:
                train_dataset = Dataset.from_dict(data_dict['train'])
                validation_dataset = Dataset.from_dict(data_dict['validation'])
                dataset_dict = {'train': train_dataset, 'validation': validation_dataset}
            # Convert the train split into a Dataset object
            else:
                # Create Dataset objects for NLI
                train_dataset = Dataset.from_dict(data_dict['train'])
                dataset_dict = {'train': train_dataset}


            if num_percentage_validation and "validation" not in data_dict:
                if not 0 < num_percentage_validation < 1:
                    raise ValueError("num_percentage_validation should be between 0 and 1")
                dataset = train_dataset.train_test_split(test_size=num_percentage_validation)
                dataset_dict = {'train': dataset['train'], 'validation': dataset['test']}

        elif task_type == 'mlm':
            # Ensure the MLM format has 'text' key
            if not isinstance(data_dict['train'], list):
                raise ValueError("For MLM, the dictionary must contain a list of 'text' under the 'train' key.")

            if "validation" in data_dict:
                train_dataset = Dataset.from_dict({"text": data_dict['train']})
                validation_dataset = Dataset.from_dict({"text": data_dict['validation']})
                dataset_dict = {'train': train_dataset, 'validation': validation_dataset}
            else:
                train_dataset = Dataset.from_dict({"text": data_dict['train']})
                dataset_dict = {'train': train_dataset}
                

            if num_percentage_validation and not "validation" in data_dict:
                if not 0 < num_percentage_validation < 1:
                    raise ValueError("num_percentage_validation should be between 0 and 1")
                dataset = train_dataset.train_test_split(test_size=num_percentage_validation)
                dataset_dict = {'train': dataset['train'], 'validation': dataset['test']}

        else:
            raise ValueError(f"Unsupported task_type '{task_type}'. Supported types: 'nli', 'mlm'.")

        # Convert labels to numeric if needed
        dataset_dict['train'] = self._convert_labels_to_numeric(dataset_dict['train'], task_type)
        if 'validation' in dataset_dict:
            dataset_dict['validation'] = self._convert_labels_to_numeric(dataset_dict['validation'], task_type)

        return DatasetDict(dataset_dict)

 
class SaveCheckpointByEpochCallback(TrainerCallback):
    """
    Callback to save the model and tokenizer at the end of each epoch during training.
 
    This callback saves the model and tokenizer state to a specified directory after each training epoch,
    allowing for periodic checkpoints of the training process.
 
    """
 
    def __init__(self, output_dir: str, tokenizer, save_checkpoint : bool, epochs_to_save : list[int], head_to_save):
        """
        Initialize the SaveCheckpointByEpochCallback.
 
        Args:
            output_dir (str): The directory where the checkpoints will be saved.
            tokenizer: The tokenizer associated with the model being trained.
        """
        self.output_dir = output_dir  # Set the directory to save the checkpoints
        self.tokenizer = tokenizer  # Set the tokenizer to be saved with the model
        self.head_to_save=head_to_save
        self.save_checkpoint=save_checkpoint
        self.epochs_to_save = epochs_to_save
 
    def on_epoch_end(self, args, state, control, model=None, **kwargs):
        """
        Save the model and tokenizer at the end of each epoch.
 
        This method is called automatically by the Trainer at the end of each epoch.
        It saves the model and tokenizer to a subdirectory named after the current epoch.
 
        Args:
            args: The training arguments.
            state: The current state of the Trainer.
            control: The current control object.
            model: The model being trained.
            **kwargs: Additional keyword arguments.
        """
        epoch = state.epoch  # Get the current epoch number
 
        if not self.output_dir:
            checkpoint_dir = f"AutoCheckpoint_{model.name_or_path}/checkpoint-epoch-{int(epoch)}"
        else:
            checkpoint_dir = f"{self.output_dir}/checkpoint-epoch-{int(epoch)}"
 
        if self.head_to_save:
            model=self.head_to_save
 
        if self.save_checkpoint:
            if not self.epochs_to_save or epoch in self.epochs_to_save:
                model.save_pretrained(checkpoint_dir)
                self.tokenizer.save_pretrained(checkpoint_dir)
 
 
 
 
 
 
class ModelTrainer:
 
    def __init__(self):
        pass
 
    def _set_nested_attribute(self, obj, attribute_string: str, value):
        """
        Set the value of a nested attribute in an object.
 
        This method sets the value of a nested attribute (e.g., "layer1.layer2.weight") in an object.
 
        Args:
            obj: The object containing the nested attribute.
            attribute_string (str): A string representing the nested attribute (e.g., "layer1.layer2.weight").
            value: The value to set for the specified nested attribute.
        """
        attrs = attribute_string.split('.')  # Split the attribute string into individual attributes
        current_obj = obj
        # Traverse the attribute hierarchy except for the last attribute
        for attr in attrs[:-1]:
            current_obj = getattr(current_obj, attr)  # Get the nested object
        setattr(current_obj, attrs[-1], value)  # Set the final attribute value
 
    def _get_nested_attribute(self, obj, attribute_string: str):
        """
        Get the value of a nested attribute from an object.
 
        This method retrieves the value of a nested attribute (e.g., "layer1.layer2.weight") from an object.
 
        Args:
            obj: The object containing the nested attribute.
            attribute_string (str): A string representing the nested attribute (e.g., "layer1.layer2.weight").
 
        Returns:
            The value of the specified nested attribute.
        """
        attributes = attribute_string.split(".")  # Split the attribute string into individual attributes
        layer_obj = obj
        # Traverse the attribute hierarchy
        for attribute_name in attributes:
            layer_obj = getattr(layer_obj, attribute_name)  # Get the nested object
        return layer_obj  # Return the final attribute value    
 
    def fix_model_embedding_layer(self, nli_model):
        # Print the original type_vocab_size
        print("Old type_vocab_size:", nli_model.config.type_vocab_size)

        nli_model.config.type_vocab_size = 2

        # Locate the token type embedding layer
        old_token_type_embeddings = nli_model.bert.embeddings.token_type_embeddings
        hidden_size = old_token_type_embeddings.embedding_dim

        # Create a new embedding layer with 2 rows (for two segment types) and the same hidden size
        new_token_type_embeddings = nn.Embedding(2, hidden_size)

        # Initialize the new embeddings:
        # Copy the existing embedding (row 0) for both token type 0 and token type 1.
        with torch.no_grad():
            new_token_type_embeddings.weight.data[0] = old_token_type_embeddings.weight.data[0].clone()
            new_token_type_embeddings.weight.data[1] = old_token_type_embeddings.weight.data[0].clone()

        nli_model.bert.embeddings.token_type_embeddings = new_token_type_embeddings

        print("New type_vocab_size:", nli_model.config.type_vocab_size)

 
 
    def _build_training_args(self, validate, per_device_train_batch_size, num_train_epochs, learning_rate, logging_dir, output_dir, overwrite_output_dir, save_strategy, per_device_eval_batch_size=None, evaluation_strategy='no'):
        training_args_dict = {
            "per_device_train_batch_size": per_device_train_batch_size,
            "num_train_epochs": num_train_epochs,
            "learning_rate": learning_rate,
            "logging_dir": logging_dir,
            "output_dir": output_dir,
            "overwrite_output_dir": overwrite_output_dir,
            "save_strategy": save_strategy,
            "logging_strategy":"epoch",
            "max_grad_norm": 1.0,  # Gradient Clipping
            "fp16": True  # Mixed Precision Training (FP16)
        }
 
        if validate:
            training_args_dict["per_device_eval_batch_size"] = per_device_eval_batch_size
            training_args_dict["evaluation_strategy"] = evaluation_strategy
 
        return TrainingArguments(**training_args_dict)
 
 
 
    def _build_trainer(self, model, args, train_dataset, data_collator, compute_metrics, callbacks, task_type, eval_dataset=None, preprocess_logits_for_metrics=None):
        trainer_args = {
            "model": model,
            "args": args,
            "train_dataset": train_dataset,
            "data_collator": data_collator,
            "compute_metrics": compute_metrics,
            "callbacks": callbacks,
        }
        if eval_dataset:
            trainer_args['eval_dataset']=eval_dataset
        if task_type=="mlm":
            trainer_args['preprocess_logits_for_metrics']=preprocess_logits_for_metrics
        return Trainer(**trainer_args)
 
 
 
    def init_head(self, uninitialized_head : AutoModelForMaskedLM, initialized_head : AutoModelForMaskedLM, layers_to_init : list[str]):
        model_name = uninitialized_head.base_model.config._name_or_path   
        print(f"===================================Copying layers weights and biases to {model_name} model===========")
        # this is done to copy the whole layer and not just an attribute of it, for example, at first we get: "vocab_transform.weight", and I want to access the whole layer "vocab_transform"
        layers_to_init = list(set([".".join(layer.split(".")[:-1]) for layer in layers_to_init]))
        for init_layer_name in layers_to_init:
            if "." in init_layer_name: # if there are iterative nested attributes, for example: lm_head.decoder
 
                layer_obj = self._get_nested_attribute(initialized_head, init_layer_name) 
                self._set_nested_attribute(uninitialized_head, init_layer_name, layer_obj)
 
            else:           
                setattr(uninitialized_head, init_layer_name, getattr(initialized_head, init_layer_name))
            print(f"The {init_layer_name} layer was copied from the initialized head!")            
        print("===================================Done copying layers weights and biases===================================")
 
 
 
 
    def _preprocess_logits_for_metrics_mlm(self, logits, labels):
        if isinstance(logits, tuple):
            # Depending on the model and config, logits may contain extra tensors,
            # like past_key_values, but logits always come first
            logits = logits[0]
        return logits.argmax(dim=-1)
 
 
    def _compute_metrics_mlm(self, eval_pred):
        predictions, labels = eval_pred
        #predictions = logits.argmax(-1)
        metric = load_metric("accuracy")
 
        predictions = predictions.reshape(-1)
        labels = labels.reshape(-1)
        # Convert predictions and labels to lists
        mask = labels != -100       
        labels = labels[mask]
        predictions = predictions[mask]
 
        return metric.compute(predictions=predictions, references=labels)
 
 
    def _compute_metrics_nli(self, p: EvalPrediction):
        preds = p.predictions[0] if isinstance(p.predictions, tuple) else p.predictions
        preds = np.argmax(preds, axis=1)
        metric = load_metric("accuracy")
        result = metric.compute(predictions=preds, references=p.label_ids)
        if len(result) > 1:
            result["combined_score"] = np.mean(list(result.values())).item()
        return result
 
 
    def _freeze_base_model(self, model, freeze_base):
        for param in model.base_model.parameters():
            param.requires_grad = not freeze_base      
 
 
    def _get_min_sequence_length(self, tokenizer, dataset, task_type):
        model_max_tokens = tokenizer.model_max_length                                        
        def find_longest_sequence(dataset, tokenizer):
            max_length = 0
            for sample in dataset:
                if task_type=="nli":
                    inputs = tokenizer(sample['premise'], sample['hypothesis'], truncation=False)
                elif task_type=="mlm":
                    inputs = tokenizer(sample['text'], truncation=False)
                seq_length = len(inputs['input_ids'])
                if seq_length > max_length:
                    max_length = seq_length
            return max_length
 
        train_max_length = find_longest_sequence(dataset['train'], tokenizer)
        if 'validation' in dataset.keys():
            validation_max_length = find_longest_sequence(dataset['validation'], tokenizer)
            longest_sequence = max(train_max_length, validation_max_length)
        else:
            longest_sequence = train_max_length
 
        training_model_max_tokens = min(model_max_tokens, longest_sequence)
        return training_model_max_tokens    
 
 
    def _train_mlm(self, model, tokenizer, dataset: Union[str, DatasetDict, dict], val_dataset,shuffle_dataset, batch_size,num_percentage_validation,num_samples_train, num_samples_validation, num_epochs, learning_rate, save_checkpoint, checkpoint_path, epochs_to_save, training_model_max_tokens, head_to_save, freeze_base):
        task_type="mlm"
        self.data_loader = DataLoader()
 
        def preprocess_function(dataset):
            return tokenizer(dataset['text'], truncation=True, padding=True, max_length=training_model_max_tokens)

 
        if isinstance(dataset, str):
            if not dataset.endswith(".csv"):
                raise ValueError("The dataset must be a path to a CSV file.")
            dataset = self.data_loader._load_csv_data(dataset_path=dataset, task_type="mlm", num_percentage_validation=num_percentage_validation, val_dataset=val_dataset)
 
        elif isinstance(dataset, DatasetDict):
            pass
        #     if val_dataset:
        #         raise ValueError("Include the validation set in the `DatasetDict`!")
 
        elif type(dataset) is dict:
            if val_dataset:
                raise ValueError("Include the validation set in the dictionary!")
            if  not (all(isinstance(item, str) for item in dataset['train']) and (all(isinstance(item, str) for item in dataset['validation']) if 'validation' in dataset.keys() else True)):
                raise ValueError("The data must be strings contained in a list!")
 
            dataset = self.data_loader._prepare_dict_dataset(data_dict=dataset, task_type=task_type, num_percentage_validation=num_percentage_validation, val_dataset=val_dataset)
 
        else:
            raise TypeError("Unsupported dataset type. Please provide a path to a CSV file, a DatasetDict, or a dictionary of data.")
 
        if shuffle_dataset:
            for key in dataset.keys():
                dataset[key] = dataset[key].shuffle(seed=42)
                
        if num_samples_train:
            dataset['train'] = dataset['train'].select(range(num_samples_train))

        if num_samples_validation and "validation" in dataset:
            dataset['validation'] = dataset['validation'].select(range(num_samples_validation)) 
 
        self.data_loader._print_dataset_status(dataset, task_type=task_type)
 
        if not training_model_max_tokens:
            training_model_max_tokens = self._get_min_sequence_length(tokenizer=tokenizer, dataset=dataset, task_type=task_type)    
 
 
        tokenized_dataset = dataset.map(preprocess_function, batched=True)
 
        train_sampled_dataset=tokenized_dataset["train"]
 
        validate = False
        validation_sampled_dataset=None
        if "validation" in tokenized_dataset:
            validation_sampled_dataset=tokenized_dataset["validation"]
            validate=True
 
        data_collator = DataCollatorForLanguageModeling(tokenizer=tokenizer, mlm_probability=0.15)
 
        if freeze_base:
            self._freeze_base_model(model=model, freeze_base=freeze_base)
        
        training_args = self._build_training_args(validate=validate,per_device_train_batch_size=batch_size, per_device_eval_batch_size=batch_size, num_train_epochs=num_epochs, learning_rate=learning_rate, evaluation_strategy="epoch", logging_dir="./mlm_training/logs/logging_mlm", output_dir="./mlm_training/output", overwrite_output_dir = True, save_strategy="no")
        trainer = self._build_trainer(model=model, args=training_args, train_dataset=train_sampled_dataset, eval_dataset=validation_sampled_dataset, data_collator=data_collator,task_type=task_type, compute_metrics=self._compute_metrics_mlm,preprocess_logits_for_metrics=self._preprocess_logits_for_metrics_mlm, callbacks=[SaveCheckpointByEpochCallback(checkpoint_path, tokenizer, save_checkpoint, epochs_to_save, head_to_save=head_to_save)])
 
        # Train the model
        trainer.train()
        return model
 
 
 
    def _train_nli(self, model, tokenizer, dataset : Union[str, DatasetDict, dict], val_dataset, label2_id, num_samples_train, num_samples_validation, num_percentage_validation, shuffle_dataset, batch_size, num_epochs, learning_rate,save_checkpoint , checkpoint_path, epochs_to_save, training_model_max_tokens, head_to_save, freeze_base):
        
        if hasattr(model.config, 'type_vocab_size') and model.config.type_vocab_size < 2:
            self.fix_model_embedding_layer(model)
        
        task_type="nli"
        self.data_loader = DataLoader(label2_id=label2_id)
        # Tokenize the combined dataset
        def preprocess_function(dataset):
            return tokenizer(dataset['premise'], dataset['hypothesis'], padding="max_length", truncation=True, max_length=training_model_max_tokens, return_tensors="pt")  
 
        if isinstance(dataset, str):
            if not dataset.endswith(".csv"):
                raise ValueError("The dataset must be a path to a CSV file.")
            dataset = self.data_loader._load_csv_data(dataset_path=dataset, task_type="nli", num_percentage_validation=num_percentage_validation, val_dataset=val_dataset)
 
        elif isinstance(dataset, DatasetDict):
            pass
        #     if val_dataset:
        #         raise ValueError("Include the validation set in the DatasetDict object!")
 
        elif type(dataset) is dict:
            if val_dataset:
                raise ValueError("Include the validation set in the dictionary!")
            if not (all(isinstance(item, str) for item in dataset['train']) and (all(isinstance(item, str) for item in dataset['validation']) if 'validation' in dataset.keys() else True)):
                raise ValueError("The data must be strings contained in a list!")
 
            dataset = self.data_loader._prepare_dict_dataset(data_dict=dataset, task_type=task_type, num_percentage_validation=num_percentage_validation, val_dataset=val_dataset)
 
        else:
            raise TypeError("Unsupported dataset type. Please provide a path to a CSV file, a DatasetDict, or a dictionary of data.")

        if shuffle_dataset:
            for key in dataset.keys():
                dataset[key] = dataset[key].shuffle(seed=42)
                
        if num_samples_train:
            dataset['train'] = dataset['train'].select(range(num_samples_train))

        if num_samples_validation and "validation" in dataset:
            dataset['validation'] = dataset['validation'].select(range(num_samples_validation))        
 
        self.data_loader._print_dataset_status(dataset, task_type=task_type)
 
        if not training_model_max_tokens:
            training_model_max_tokens = self._get_min_sequence_length(tokenizer=tokenizer, dataset=dataset, task_type=task_type)    
 
        print(f"training_model_max_tokens: {training_model_max_tokens}")
        tokenized_dataset = dataset.map(preprocess_function, batched=True)
 
        train_sampled_dataset=tokenized_dataset["train"]
        validate = False
        validation_sampled_dataset=None
        if "validation" in tokenized_dataset:
            validation_sampled_dataset=tokenized_dataset["validation"]
            validate = True
 
        data_collator = DataCollatorWithPadding(tokenizer=tokenizer)
        if freeze_base:
            self._freeze_base_model(model=model, freeze_base=freeze_base)        
 
        training_args = self._build_training_args(validate=validate,per_device_train_batch_size=batch_size, per_device_eval_batch_size=batch_size, num_train_epochs=num_epochs, learning_rate=learning_rate, evaluation_strategy="epoch", logging_dir="./nli_training/logs/logging_nli", output_dir="./nli_training/output", overwrite_output_dir = True, save_strategy="no")
        trainer = self._build_trainer(model=model, args=training_args, train_dataset=train_sampled_dataset, eval_dataset=validation_sampled_dataset, data_collator=data_collator, task_type=task_type, compute_metrics=self._compute_metrics_nli, callbacks=[SaveCheckpointByEpochCallback(checkpoint_path, tokenizer, save_checkpoint, epochs_to_save, head_to_save=head_to_save)])
 
 
        # Train the model
        trainer.train()
        return model
 
 
    def get_non_base_layers(self, model):
 
        all_layers = list(model.state_dict().keys())
        base_layers = list(model.base_model.state_dict().keys())
        head_layers=[]
        for layer in all_layers:
            if ".".join(layer.split(".")[1:]) not in base_layers: # when looping over the layers of the base model we want to remove the prefix of the layer which is the name of the model, hence the ".".join(layer.split(".")[1:])
                head_layers.append(layer)
 
        return head_layers
 
 
    def attach_head_to_model(self, head1, head2, model_identifier : str):       
        setattr(head1, model_identifier, getattr(head2 ,model_identifier))
 
 
 
    def train_head(self, model, tokenizer, dataset, label2_id=None, nli_head=False, mlm_head=False, 
                   model_to_copy_weights_from=None,num_samples_train=None, num_samples_validation=None, num_percentage_validation=None,
                   shuffle_dataset=False, val_dataset=None, batch_size=16, num_epochs=10, learning_rate=2e-5,training_model_max_tokens = None, freeze_base = False, copy_weights=False, 
                   save_checkpoint=False, checkpoint_path=None, epochs_to_save=None, head_to_save=None):
 
        model_name = model.base_model.config._name_or_path
        if (num_percentage_validation and val_dataset):
            raise ValueError("Cannot specify `num_percentage_validation` when validation set is available!")

        if (num_samples_validation and not val_dataset and not num_percentage_validation):
            raise ValueError("`val_dataset` or `num_percentage_validation` must be specified to use `num_samples_validation`!")
        
        if (nli_head and not label2_id):
            raise ValueError("`label2_id` must be specified for an NLI task!")

        if  (not nli_head and not mlm_head) or (nli_head and mlm_head): # if both false or both true
            raise ValueError("You must have one head (nli_head or mlm_head) set to True at a time.")
 
        if copy_weights:
            if not model_to_copy_weights_from:
                raise ValueError("Please pass in a model (model_to_copy_weights_from=?) to load the initialized layers from!")
 
 
            get_initialized_layers = self.get_non_base_layers(model_to_copy_weights_from)
            get_uninitialized_layers = self.get_non_base_layers(model)
            if sorted(get_uninitialized_layers)!=sorted(get_initialized_layers):
                raise ValueError(f"Models architecture are not equal, make sure that {model_to_copy_weights_from.base_model.config._name_or_path} head layers are the same as {model_name}'s")
            self.init_head(model, model_to_copy_weights_from, get_uninitialized_layers)
 
 
        if nli_head:
            print(f"Detected {model_name} with an NLI head...")
            self._train_nli(model=model, tokenizer=tokenizer, dataset=dataset, num_percentage_validation=num_percentage_validation, num_samples_train=num_samples_train, num_samples_validation=num_samples_validation, shuffle_dataset=shuffle_dataset,val_dataset=val_dataset,label2_id=label2_id, batch_size=batch_size, num_epochs=num_epochs, learning_rate=learning_rate,save_checkpoint=save_checkpoint, checkpoint_path=checkpoint_path, epochs_to_save=epochs_to_save, training_model_max_tokens=training_model_max_tokens, head_to_save=head_to_save, freeze_base=freeze_base)
 
        elif mlm_head:
            print(f"Detected {model_name} with an MLM head...")
            self._train_mlm(model=model, tokenizer=tokenizer, dataset=dataset, num_percentage_validation=num_percentage_validation, num_samples_train=num_samples_train, num_samples_validation=num_samples_validation, shuffle_dataset=shuffle_dataset,val_dataset=val_dataset, batch_size=batch_size, num_epochs=num_epochs, learning_rate=learning_rate,save_checkpoint=save_checkpoint, checkpoint_path=checkpoint_path, epochs_to_save=epochs_to_save, training_model_max_tokens=training_model_max_tokens, head_to_save=head_to_save, freeze_base=freeze_base)


import torch
import pandas as pd
import numpy as np
import torch
import time
from pprint import pprint
from transformers import AutoModelForMaskedLM, AutoTokenizer
from transformers import pipeline
from transformers import PreTrainedModel
from transformers import PreTrainedTokenizer
import scipy
import sklearn as sk
import itertools
from ..qabstract.qabstract import *
from ..qabstract.qabstract import SCALE, DIMENSIONS, FILTER, IDXSELECT, _filter_data_frame

from transformers import BertTokenizer, BertForNextSentencePrediction
import torch


class NextSentencePredictionPipeline():
  def __init__(self, model_name='bert-base-uncased', device=0):
    self.tokenizer = BertTokenizer.from_pretrained(model_name)
    self.model = BertForNextSentencePrediction.from_pretrained(model_name, return_dict=True)
    self.model.to(device)

  def __call__(self, sentence_pairs, device=0):
    results = []
#     print(sentence_pairs)
#     for prompt, next_sentence in sentence_pairs:
    prompt = sentence_pairs[0]
    next_sentence = sentence_pairs[1]
    encoding = self.tokenizer(prompt, next_sentence, return_tensors='pt').to(device)
    next_sentence_label = torch.LongTensor([1]).to(device)
    outputs = self.model(**encoding, labels=next_sentence_label)
#     outputs = self.model(**encoding, next_sentence_label=torch.LongTensor([1]))
    logits = outputs.logits
    results.append(logits)
    return torch.vstack(results).detach()


class QNSP(QABSTRACT):
    # Static property QNSP._qregister is defined after this class
    # It will contain one instance (the last one) of every subtype of QNSP.
    # This will allow to automatically register questions and reuse them with various models when the time comes.

    def __init__(self, next_sentence: str, prompt: str, dimensions:DIMENSIONS = {},
                model: pipeline = None,
                 p=None,
                 index=None,
                 scale='intensifier',
                 descriptor = {}):
        super().__init__(dimensions, model, p, index, scale, descriptor)
#         self.nsp = model
        self._next_sentence = next_sentence
        self._prompt = prompt
        self._descriptor['query'] = prompt+"->"+next_sentence
        QNSP._qregister[self.__class__.__name__] = self

    def ans_logits(self, result):
        return torch.nn.functional.softmax(result, dim=1)[:,0]

    def run(self, model=None):
        super().run(model)
        T = time.time()
        coo = []
        p = []  
        for kmap,kcoo in zip(self._keywords_map,self._keywords_grid_idx):
            prompt = self._prompt.format_map(kmap)
            next_sentence = self._next_sentence.format_map(kmap)
            val = torch.softmax(self.model([prompt, next_sentence]), dim=1)[0]
            prob = val[0].item()
            p.append(torch.Tensor([prob]).squeeze().cpu().item())
            coo.append(kcoo)

        coo = torch.stack(coo).T
        assert torch.all(torch.eq(coo.T, self._keywords_grid_idx))

        self._pdf["P"] = p
#         print(self._pdf)
        self._T = time.time() - T
        self.result = self
#         print(self.result)
        return self.result


from typing import Dict

# The static property QNSP._qregister will contain one instance (the last one) of every subtype of QNSP.
# This will allow to automatically registering questions and reusing them with various models when the time comes.
QNSP._qregister: Dict[str, QNSP] = {}



import torch
import pandas as pd
import numpy as np
from pprint import pprint
from transformers import AutoModelForMaskedLM, AutoTokenizer
from transformers import pipeline
from transformers import PreTrainedModel
from transformers import PreTrainedTokenizer
import scipy
import itertools
import re

from ..qabstract.qabstract import *
from ..qabstract.qabstract import SCALE, DIMENSIONS, FILTER, IDXSELECT, _filter_data_frame


class QMLM(QABSTRACT):    
    def __init__(self,
                 template:str,
                 dimensions:DIMENSIONS = {},
                 model:pipeline = None,
                 p=None,
                 index=None,
                 scale='intensifier',
                 descriptor = {}):
        super().__init__(dimensions, model, p, index, scale, descriptor=descriptor)
        
        self._index = index
        self._scale = scale
        self._descriptor['query'] = template
        self._template = template
        QMLM._qregister[self.__class__.__name__] = self
        
    
    def ans_logits(self, result):
        ans = [r['token_str'] for r in result]
        prob = torch.tensor([r['score'] for r in result])
        return dict(zip(ans, prob))

    
    def mlm_ans(mlm_esult):
        return sorted([r['token_str'] for r in mlm_esult])


    def chain_prob(self, query, word, j=0):
        tokens = self.model.tokenizer(word, add_special_tokens = False)
        token_str = self.model.tokenizer.convert_ids_to_tokens(tokens['input_ids'])
        token_count = len(token_str)
        mask_token = self.model.tokenizer.mask_token  

        if token_count > 1 and token_str[1].startswith('##'):                    
            mask = token_count*mask_token
        else:
            mask = ' '.join(token_count*[mask_token])

        mask_num = query.count(mask_token) 
        start_index = [m.start() for m in re.finditer(re.escape(mask_token), query)][j]
        query = query[:start_index] + mask + query[start_index + len(mask_token):]

        ans = self.model(query, targets=token_str[:1])

        if token_count > 1:
            probs = [ans[j][0]["score"]]
            for i in range(1, token_count): 
                start_index = [m.start() for m in re.finditer(re.escape(mask_token), query)][j] 
                token = token_str[i - 1]
                mask_len = len(mask_token)
                query = query[:start_index] + token + query[start_index + mask_len:]
                ans = self.model(query, targets=token_str[i:i+1])

                if mask_num == 1:
                    if token_count - i == 1:
                        probs.append(ans[j]["score"])
                    else:
                        probs.append(ans[j][0]["score"])
                else:
                    probs.append(ans[j][0]["score"])
        else:
            if mask_num == 1:
                probs = [ans[0]["score"]]
            else:
                probs = [ans[j][0]["score"]]
        prob = np.prod(probs)
        return prob
    
    def run(self, model= None):
        super().run(model)
        tokens_to_ids_dict = {}
        str_to_tokens_dict = {}
        mask_token = self.model.tokenizer.mask_token
        for key, value in self._keywords.items():
            for i in value:
                tokens = self.model.tokenizer(i, add_special_tokens = False)
                token_str = self.model.tokenizer.convert_ids_to_tokens(tokens['input_ids'])
                tokens_to_ids_dict[tuple(token_str)] = tokens['input_ids'];
                str_to_tokens_dict[i] = tuple(token_str)
        T = time.time()
        coo = []
        p = []
        for kmap,kcoo in zip(self._keywords_map,self._keywords_grid_idx):
            dimensions_probs = []
#             dimension_name = self._index[0]
            dimension_name = self._scale
            dimension_value = kmap[dimension_name]
            index_value = kmap[self._index[0]]
            masked_kmap = kmap.copy()
            masked_kmap[dimension_name] = self.model.tokenizer.mask_token 
            prob = self.chain_prob(self._template.format_map(masked_kmap), dimension_value, 0)
            
            dimensions_probs.append(prob)
            kmap[dimension_name] = dimension_value
                
            # avg_probs_dimensions is the averege of the propability of the sentence
            avg_prob_dimensions = np.mean(dimensions_probs)
            p.append(avg_prob_dimensions)
            coo.append(kcoo)

        coo = torch.stack(coo).T
        assert torch.all(torch.eq(coo.T, self._keywords_grid_idx))

        self._pdf["P"] = p
        
        self._T = time.time() - T
        self.result = self
        return self.result


from typing import Dict

# The static property QCOLA._qregister will contain one instance (the last one) of every subtype of QCOLA.
# This will allow to automatically registering questions and reusing them with various models when the time comes.
QMLM._qregister: Dict[str, QMLM] = {}

class QQQ():
    pass

class QQW():
    pass






import torch
import gc
import numpy as np
import pandas as pd
from matplotlib import *
import matplotlib.image as im
import matplotlib.pyplot as plt
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
from matplotlib.patches import Patch
from pathlib import Path
from IPython.core.interactiveshell import InteractiveShell
InteractiveShell.ast_node_interactivity = "all"
from abc import *
from typing import *
from numbers import Number
from sentence_transformers import SentenceTransformer, util
from transformers import pipelines, AutoModel, Pipeline
import pingouin as pg
from itertools import chain
from abc import ABC, abstractmethod
from jinja2 import Template
import markdown
from datetime import datetime
import pytz
from tqdm import tqdm
from reportlab.platypus import Paragraph
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib.units import inch
from reportlab.lib import colors
from reportlab.lib.pagesizes import landscape, letter
from reportlab.pdfgen import canvas
from reportlab.platypus import Paragraph
from reportlab.lib.styles import getSampleStyleSheet
import csv
# from qlatent.qabstract.qabstract import *
# change for the path of qlatent files
from qlatent.qabstract.qabstract import *
from huggingface_hub import HfApi
hf_api = HfApi()


device = 0 if torch.cuda.is_available() else -1
print(device)

FILTER = Dict[str,Collection[str]]


class Questionnaire:
    """
       Class representing a Questionnaire.
       The class has the following parameters:
           name (str): The name of the questionnaire.
           num_of_questions (int): the number of the total questions of the questionnaire.
           factors (List(str)): the factors of the questionnaire.
           factor_grouping (Dict[str -> set(str)]): Used for grouping sub factor into one factor.
           questions (dict[int -> Qabstract]): the questions of the questionnaire are stored here, mapped via their ordinal number.
    """

    def __init__(self, name : str, num_of_questions : int, factors : List[str], factor_grouping:Dict[str,Set[str]]=None, full_name:str=None):
        # input validation
        if not isinstance(name, str):
            raise TypeError(f"Questionnaire name must be a string, got type {type(name)} instead.")
        if full_name is None:
            full_name = name
        elif not isinstance(full_name, str):
            raise TypeError(f"Questionnaire full name must be a string, got type {type(name)} instead.")
        if not isinstance(num_of_questions, int):
            raise TypeError(f"Questionnaire num_of_questions must be an int, got type {type(num_of_questions)} instead.")
            if (num_of_questions < 1):
                raise ValueError(f"Questionnaire num_of_questions must be larger than one (1), got {num_of_questions} instead.")
        if not (isinstance(factors, list) and all(isinstance(factor, str) for factor in factors)):
            raise TypeError("Questionnaire factors must be a list of strings.")
        
        self.name = name
        self.full_name = full_name
        self.num_of_questions = num_of_questions
        self.factors = {factor : set() for factor in factors}
        self.set_factor_grouping(factor_grouping)
        self.questions = {}
        
        
    def set_factor_grouping(self, factor_grouping:Dict[str,Set[str]]=None):
        if factor_grouping is None:
            self._factor_grouping = {factor : {factor,} for factor in self.factors}
        elif not (isinstance(factor_grouping, dict) and all(isinstance(k, str) and isinstance(v, set) and all(isinstance(i, str) and i in self.factors for i in v)  for k, v in factor_grouping.items())):
            raise TypeError("Questionnaire factor_split must be a dictionary mapping string (factor grouping) to set of strings (current factors).")
        else: # user provided valid factor_grouping
            self._factor_grouping = factor_grouping
            self._update_factor_grouping()
            
                
    def _update_factor_grouping(self):
        # make sure the factor grouping addresses all of the current factors in the questionnaire
        factors_covered_in_grouping = set()
        for factors in self._factor_grouping.values():
            factors_covered_in_grouping.update(factors)

        for factor in self.factors.keys(): # add missing factors to factor_grouping
            if factor not in factors_covered_in_grouping: # factor not represented in factor grouping
                    self._factor_grouping[factor] = {factor,}
        
        for factors in self._factor_grouping.values(): # remove unnecessary factors from _factor_grouping
            factors.intersection_update(self.factors.keys())
            
        # Delete factor groups that hold empty sets
        self._factor_grouping = {group_name: factors for group_name, factors in self._factor_grouping.items() if factors}
                        
        
    @classmethod
    def create_questionnaire_from_questions(cls, questions : List[QABSTRACT]):
        if not isinstance(questions, list):
            raise ValueError("questions parameter must must be of type list.")
        if len(questions) == 0:
            raise ValueError("questions parameter must contain at least one question.")
            
        # create questions instances
        questions = [q() for q in questions]
                
        # Sample questionnaire name from the first question
        questionnaire_name = questions[0]._descriptor["Questionnair"]
        
        obj = cls(name = questionnaire_name,
                   num_of_questions = len(questions),
                   factors = [],
                   factor_grouping=None,
                   full_name=None)
        
        for q in questions:
             obj.add_question(q)
        
        return obj
    
     
    def is_complete(self) -> bool:
        return len(self.questions) == self.num_of_questions

    
    def __getitem__(self, index):
        return self.questions[index]
        

    def __len__(self):
        return len(self.questions)
        
    
    def add_question(self, q: QABSTRACT):
        """ adds a question object into the Questionnaire, and updates the factors dictionary accordingly """
        question_factor = q._descriptor['Factor']
        question_ordinal = q._descriptor['Ordinal']
        self.questions[question_ordinal] = q
        if len(self) > self.num_of_questions:
            self.num_of_questions = len(self)
        if question_factor in self.factors:
            self.factors[question_factor].add(question_ordinal)
        else:
            self.factors[question_factor] = {question_ordinal,}
            self._update_factor_grouping()
            #print(f"Added a new factor <{question_factor}> to the questionnaire {self.name} based on question with ordinal <{question_ordinal}>")
            
            


    def remove_question_by_ordinal(self, question_ordinal:int):
        """ Removes a question object from the questionnaire """
        if question_ordinal in self.questions:
            q = self.questions[question_ordinal]
            question_factor = q._descriptor['Factor']
            del self.questions[question_ordinal]
            self.factors[question_factor].remove(question_ordinal)
            self._update_factor_grouping()
        else:
            raise KeyError(f"Questionnaire does not contain a question with ordinal number <{question_ordinal}>")
            
            
    def remove_question_by_object(self, q:QABSTRACT):
        """ Removes a question object from the questionnaire """
        
        # find the object in self.questions
        for ordinal, obj in self.questions.items():
            if q is obj: # exact match for the same object
                del self.questions[ordinal]
                factor = q._descriptor['Factor']
                self.factors[factor].remove(ordinal)
                self._update_factor_grouping()
                return
        
        raise ValueError(f"Questionnaire does not contain the given question object <{q}>")
        
    
    def run(self,
            pipelines : List[Pipeline],
            questions_ordinals:List[int]=None,
            result_path=Path('./results/result.csv'),
            softmax:List[str]=[],
            filters:Dict[str,dict]={},
            merge_filtered_positiveonly=False,
            ):
        """ Runs given models on specified questions with pipeline template defined inside pipeline_builder
            
            pipelines : List[Pipeline] A list of HuggingFace pipelines to evaluate on a questionnaire.
            questions_ordinals:List[int] A list of integers representing the questions to run the models on, if left empty than the models will run over all the questions of the questionnaire.
            result_path The full path to save the result csv to, defualt location is './results/result.csv'
            softmax:List[str] List of strings representing which dimentions to apply softmax on, order of item matters.
            filters:Dict[str,dict] filters to apply to a questionnaire, each filter will result in additional logging for every question.
        """
        if questions_ordinals is None:
            questions_ordinals = list(self.questions.keys())
        for pipeline in pipelines:
            self.run_pipeline_on_questions(pipe=pipeline,
                                        questions_ordinals=questions_ordinals,
                                        softmax=softmax,
                                        filters=filters,
                                        result_path=result_path,
                                        merge_filtered_positiveonly=merge_filtered_positiveonly
                                        )


        # self.calc_correlations()
        # self.report()
        
        
            
    def run_pipeline_on_questions(self, questions_ordinals, pipe, result_path, softmax, filters: Dict[str, dict], merge_filtered_positiveonly):
        print(f"\tEvaluating {self.name} questionnaire on {pipe.model_identifier}: ", flush=True)
        commits = hf_api.list_repo_commits(repo_id=pipe.model_identifier)
        last_commit = commits[0]
        last_commit_hash = last_commit.commit_id
        last_commit_date = last_commit.created_at.strftime("%d-%m-%Y %H:%M:%S") + " UTC"
        
        for ordinal in tqdm(questions_ordinals):
            q = self.questions[ordinal]
            q = QCACHE(q)
            
            # Apply the softmax logic
            if (softmax is None) or (isinstance(softmax, list) and len(softmax) == 0):
                qsf = QPASS(q, descupdate={'softmax': ''})
            else:
                qsf = QSOFTMAX(q, dim=softmax)
            
            # Apply the filter logic
            if (filters is None) or (isinstance(filters, dict) and len(filters) == 0):
                filters = {'unfiltered': lambda q: {}}
            for filtername, filter_function in filters.items():                
                qsf_f = QFILTER(qsf, filter=filter_function(qsf), filtername=filtername)
                # Run the pipeline on the question
                qsf_f.run(pipe)
                # Retrieve attributes after the run method has populated the necessary data
                attributes = self.question_attributes(qsf_f, last_commit_hash, last_commit_date)
                attributes = list(attributes.items())
                attributes=dict(attributes)
                # Create and save attributes as a DataFrame
                attributes_df = pd.DataFrame([attributes])  # Wrap in list

                # Convert to list of items
                self.append_df_to_csv(attributes_df, result_path)
        if merge_filtered_positiveonly:
            self.merge_unfiltered_into_positive_only(result_path)

    def question_attributes(self, q, last_commit_hash, last_commit_date):
        descriptor = q._descriptor
        now = datetime.now(pytz.UTC)
        date_of_logging = now.strftime('%d-%m-%Y %H:%M:%S %Z')
        model_id = q.model.model_identifier if q.model.model_identifier else "no model detected"
        silhouette_table_structure = self.classify_heatmap_pattern(q._pdf, frequency_weights=q._dimensions["frequency"], index_weights=q._dimensions["index"])
        attributes = {
            "questionnaire" : descriptor['Questionnair'],
            "factor" : descriptor['Factor'],
            "ordinal" : descriptor['Ordinal'],
            'model' : model_id,
            "original" : descriptor['Original'],
            "translation_method" : None, # This value is set inside the inhereting class
            'query' : descriptor['query'],
            'scale' : q._scale,
            'index' : q._index,
            'dimensions' : q._dimensions,
            'filter' : descriptor['filter'],
            'softmax' : descriptor['softmax'],            
            'mean_score' : q.mean_score(),
            'silhouette_score' : Questionnaire.silhouette_score(q),
            'silhouette_table_structure' : silhouette_table_structure,
            "last_commit_hash" : last_commit_hash,
            "commit_date" : last_commit_date,
            "model_version_id" : model_id+"_"+last_commit_hash,
            "logging_date" : date_of_logging,
        }        
        return attributes
    
    
    @staticmethod
    def silhouette_score(q):
        try:
            silhouette_score= q.internal_consistency(
                measure='silhouette_score',
                metric='correlation',
                index=q._index,
                scale=q._scale,
            )
            return silhouette_score
        except Exception as e:
            print(f"Failed to calculate silhouette score for question with ordinal {q._descriptor['Ordinal']} with filter {q._descriptor['filter']}, Error: {e}")
            return None



    def merge_unfiltered_into_positive_only(self, df_path: str) -> pd.DataFrame:
        
        df = pd.read_csv(df_path, encoding='utf-8-sig')
        
        # Split data
        positive_only = df[df["filter"] == "positive_only"].copy()
        unfiltered = df[df["filter"] == "unfiltered"]

        # Merge on identifiers
        merge_keys = ["model_version_id", "ordinal"]
        merged = pd.merge(
            positive_only,
            unfiltered[merge_keys + ["silhouette_score"]],
            on=merge_keys,
            how="left",
            suffixes=("", "_from_unfiltered")
        )

        # Fill silhouette_score
        merged["silhouette_score"] = merged["silhouette_score"].fillna(merged["silhouette_score_from_unfiltered"])

        # Drop the temporary column
        merged = merged.drop(columns=["silhouette_score_from_unfiltered"])
        merged = merged.rename(columns={"silhouette_score": "silhouette_score(unfiltered)"})
        merged.to_csv(df_path, encoding="utf-8-sig", index=False)

        return merged


    def classify_heatmap_pattern(
        self,
        df: pd.DataFrame,
        score_col: str = 'P',
        frequency_weights: dict = None,
        index_weights: dict = None,
        dominance_ratio: float = 1.4,
        support_threshold: float = 0.3
    ) -> str:
        """
        Classifies a frequency × index heatmap using:
        - Original sorted order (ascending weights)
        - Semantic quadrant splits (based on polarity of weights)

        Returns:
            str: One of the pattern types:
                'diagonal', 'anti-diagonal', 'horizontal-dominant',
                'vertical-dominant', 'high_uniform', 'low_uniform', 'quarter-mixed'
        """
        matrix = df.pivot(index='index', columns='frequency', values=score_col)

        # ✅ Original sorting: ascending weight order
        if frequency_weights:
            freq_order = [k for k, _ in sorted(frequency_weights.items(), key=lambda x: x[1])]
            matrix = matrix[[f for f in freq_order if f in matrix.columns]]

        if index_weights:
            idx_order = [k for k, _ in sorted(index_weights.items(), key=lambda x: x[1])]
            matrix = matrix.loc[[i for i in idx_order if i in matrix.index]]

        matrix = matrix.dropna(axis=0, how='all').dropna(axis=1, how='all')
        vals = matrix.values
        total_mean = np.mean(vals)
        total_sum = np.sum(vals)

        # ✅ Split rows/cols by sign of weights
        neg_rows = [i for i in matrix.index if index_weights.get(i, 0) < 0]
        pos_rows = [i for i in matrix.index if index_weights.get(i, 0) >= 0]
        neg_cols = [f for f in matrix.columns if frequency_weights.get(f, 0) < 0]
        pos_cols = [f for f in matrix.columns if frequency_weights.get(f, 0) >= 0]

        # Quadrants (if any empty, use empty array)
        q1 = matrix.loc[neg_rows, neg_cols].values if neg_rows and neg_cols else np.array([])
        q2 = matrix.loc[neg_rows, pos_cols].values if neg_rows and pos_cols else np.array([])
        q3 = matrix.loc[pos_rows, neg_cols].values if pos_rows and neg_cols else np.array([])
        q4 = matrix.loc[pos_rows, pos_cols].values if pos_rows and pos_cols else np.array([])

        q1_sum, q2_sum, q3_sum, q4_sum = map(np.sum, [q1, q2, q3, q4])
        main_diag_sum = q1_sum + q4_sum
        anti_diag_sum = q2_sum + q3_sum

        q1_vals, q2_vals, q3_vals, q4_vals = map(lambda q: q.flatten() if q.size else np.array([]), [q1, q2, q3, q4])
        main_diag_vals = np.concatenate([q1_vals, q4_vals])
        anti_diag_vals = np.concatenate([q2_vals, q3_vals])
        main_diag_support = np.mean(main_diag_vals > total_mean) if main_diag_vals.size else 0
        anti_diag_support = np.mean(anti_diag_vals > total_mean) if anti_diag_vals.size else 0

        if anti_diag_sum and main_diag_sum / anti_diag_sum > dominance_ratio and main_diag_support >= support_threshold:
            return 'diagonal'
        if main_diag_sum and anti_diag_sum / main_diag_sum > dominance_ratio and anti_diag_support >= support_threshold:
            return 'anti-diagonal'

        top_vals = np.concatenate([q1_vals, q2_vals])
        bottom_vals = np.concatenate([q3_vals, q4_vals])
        top_sum, bottom_sum = q1_sum + q2_sum, q3_sum + q4_sum
        top_support = np.mean(top_vals > total_mean) if top_vals.size else 0
        bottom_support = np.mean(bottom_vals > total_mean) if bottom_vals.size else 0

        if bottom_sum and top_sum / bottom_sum > dominance_ratio and top_support >= support_threshold:
            return 'horizontal-dominant'
        if top_sum and bottom_sum / top_sum > dominance_ratio and bottom_support >= support_threshold:
            return 'horizontal-dominant'

        left_vals = np.concatenate([q1_vals, q3_vals])
        right_vals = np.concatenate([q2_vals, q4_vals])
        left_sum, right_sum = q1_sum + q3_sum, q2_sum + q4_sum
        left_support = np.mean(left_vals > total_mean) if left_vals.size else 0
        right_support = np.mean(right_vals > total_mean) if right_vals.size else 0

        if right_sum and left_sum / right_sum > dominance_ratio and left_support >= support_threshold:
            return 'vertical-dominant'
        if left_sum and right_sum / left_sum > dominance_ratio and right_support >= support_threshold:
            return 'vertical-dominant'

        quadrant_sums = [q1_sum, q2_sum, q3_sum, q4_sum]
        q_sum_range = max(quadrant_sums) - min(quadrant_sums)

        if q_sum_range < 0.1 * total_sum:
            return 'high_uniform' if total_mean >= 0.2 else 'low_uniform'

        return 'quarter-mixed'

        
    def write_df_to_csv(self, df, csv_path):
        csv_path = Path(csv_path)
        # Ensure the parent directory exists
        csv_path.parent.mkdir(parents=True, exist_ok=True)
        df.to_csv(csv_path, mode='w', header=True, index=False, encoding='utf-8-sig')
        
            
    def append_df_to_csv(self, df, csv_path):
        csv_path = Path(csv_path)
        # Ensure the parent directory exists
        csv_path.parent.mkdir(parents=True, exist_ok=True)
        
        if not csv_path.exists():  # Check if the file itself exists
            df.to_csv(csv_path, mode='w', header=True, index=False, encoding='utf-8-sig')
        else:
            df.to_csv(csv_path, mode='a', header=False, index=False, encoding='utf-8-sig')


    def calc_silhouette(self, run_path):
        run_df = pd.read_csv(run_path, encoding='utf-8-sig')

        # Calculate the mean and standard deviation of silhouette scores
        mean = run_df['silhouette_score'].mean()
        std = run_df['silhouette_score'].std()

        # Group by 'model' and calculate the count of negative silhouette scores per model
        negative_counts = run_df[run_df['silhouette_score'] < 0].groupby('model').size()

        # Calculate the average number of questions with negative silhouette scores per model
        avg_negative_per_model = negative_counts.mean()

        return mean, std, avg_negative_per_model
    
    
    def calc_alpha(self, run_path):
        run_df = pd.read_csv(run_path, encoding='utf-8-sig')

        # Drop irrelevant columns
        relevant_columns = ['questionnaire', 'factor', 'ordinal', 'model', 'mean_score']
        run_df = run_df[relevant_columns]
        
        # Add a column combining questionnaire, factor and ordinal
        run_df['Q'] = (
                    run_df['questionnaire']
                    .str.cat(run_df['factor'], sep=':')
                    .str.cat(run_df['ordinal'].astype(str), sep=':')
            )
        
        # Calculate alpha for the entire questionnaire
        questionnaire_df = run_df[run_df['questionnaire'] == self.name]
        pivot = pd.pivot_table(questionnaire_df, values='mean_score', index='model', columns='Q', aggfunc='mean')
        total_alpha = pg.cronbach_alpha(data=pivot)[0]

        # Calculate alpha per factor
        factor_alphas = {}
        for group_name, factors in self._factor_grouping.items():

            factor_group_df = questionnaire_df[questionnaire_df['factor'].isin(factors)]
            pivot = pd.pivot_table(factor_group_df, values='mean_score', index='model', columns='Q', aggfunc='mean')
            # Check the number of columns
            if pivot.shape[1] < 2:
                raise ValueError(f"Factor {group_name} -> {factors} must have at least 2 questions to calculate cronbach Alpha.")
            # Check the number of rows
            if pivot.shape[0] < 1:
                raise ValueError(f"Factor {group_name} -> {factors} must have at least 1 model observation to calculate cronbach Alpha.")
                
            factor_alphas[group_name] = pg.cronbach_alpha(data=pivot)[0]

        return total_alpha, factor_alphas
    
    
    def calc_correlations(self, run_path, correlations_path):
        run_df = pd.read_csv(run_path, encoding='utf-8-sig')
        
        # Drop irrelevant columns
        relevant_columns = ['questionnaire', 'factor', 'ordinal', 'model', 'mean_score']
        run_df = run_df[relevant_columns]
        questionnaire_df = run_df[run_df['questionnaire'] == self.name]
        
        # Iterate over factor_grouping and duplicate rows for each group
        for group_name, factors in self._factor_grouping.items():
            # If no grouping is done for a factor then skip it
            if factors == {group_name,}: continue
            # Find rows that match the current group
            group_rows = questionnaire_df[questionnaire_df['factor'].isin(factors)].copy()
            # Update the factor to the group name
            group_rows['factor'] = group_name
            # Append these rows to the questionnaire_df
            questionnaire_df = pd.concat([questionnaire_df, group_rows], ignore_index=True)
            
        pivot = pd.pivot_table(questionnaire_df, values='mean_score', index='model', columns='factor', aggfunc='mean')
        correlations = pivot.rcorr(method='spearman')
        self.write_df_to_csv(correlations, correlations_path)
        return correlations
    
    
    def calc_model_info(self, run_path):
        """
        Calculates statistics about the models run on the questionnaire.
        """
        run_df = pd.read_csv(run_path, encoding='utf-8-sig')
        model_groups = run_df.groupby('model')
        parameter_counts = {model: self.get_model_parameters(model) for model in model_groups.groups.keys()}
        valid_parameter_counts = [p for p in parameter_counts.values() if p is not None]
        num_models = len(parameter_counts)
        avg_parameters = np.mean(valid_parameter_counts) if valid_parameter_counts else 0
        std_parameters = np.std(valid_parameter_counts) if valid_parameter_counts else 0
        return num_models, avg_parameters, std_parameters
    
    
    def get_model_parameters(self, model_path):
        """
        Fetches the number of parameters for a Hugging Face model.
        """
        try:
            model = AutoModel.from_pretrained(model_path)
            return model.num_parameters()
        except Exception as e:
            print(f"Error loading model {model_path}: {e}")
            return None        
        

    ##### DASHBOARD #####
    def report(self, run_path, output_path, template_path="./report_template.md"):
        """
        Generates a report for the questionnaire analysis and saves it as a Markdown file.
    
        Args:
            run_path (str): Path to the run directory containing required data for calculations.
            output_path (str): Path where the rendered Markdown file will be saved.
            template_path (str): Path to the Markdown template file.
    
        The Markdown template file should use placeholders and loops compatible with the Jinja2 templating engine.
        Example placeholders:
            - {{ questionnaire_name }}
            - {{ avg_silhouette }}
        Example loops:
            - {% for factor in factors %} ... {% endfor %}
        """
    
        # Step 1: Read the Markdown template from a file
        with open(template_path, "r", encoding="utf-8") as file:
            markdown_template = file.read()
    
        # Step 2: Define the dynamic data
        avg_silhouette, std_silhouette, num_negative_silhouette = self.calc_silhouette(run_path)
        cronbach_alpha_total, cronbach_alpha_factors = self.calc_alpha(run_path)
        num_models, avg_parameters, std_parameters = self.calc_model_info(run_path)
        data = {
            "questionnaire_name": f"{self.full_name} ({self.name})" if self.full_name != self.name else self.name,
            "num_items": self.num_of_questions,
            "factors": self.factors.keys(),
            "avg_silhouette": f"{avg_silhouette:.4f}",
            "std_silhouette": f"{std_silhouette:.4f}",
            "num_negative_silhouette": f"{num_negative_silhouette:.4f}",
            "cronbach_alpha_total": f"{cronbach_alpha_total:.4f}",
            "cronbach_alpha_factors": {factor: f"{alpha:.4f}" for factor, alpha in cronbach_alpha_factors.items()},
            "num_models": num_models,
            "avg_parameters": f"{avg_parameters:.4f}",
            "std_parameters": f"{std_parameters:.4f}",
        }
    
        # Step 3: Render the Markdown content
        template = Template(markdown_template)
        rendered_markdown = template.render(data)
        
        # Save Markdown content to a file
        with open(output_path, "w", encoding="utf-8") as file:
            file.write(rendered_markdown)
        
        return rendered_markdown


    
    def create_dashboard(self, run_path, filename):
        """Creates the dashboard with a widescreen 16:9 aspect ratio."""
        
        # Define dimensions for 16:9 aspect ratio
        width = 16 * inch  # Width (16 units)
        height = 9 * inch  # Height (9 units)

        # Initialize canvas with custom dimensions
        c = canvas.Canvas(filename, pagesize=(width, height))

        # Dimensions
        panel_width = width * 0.25  # 25% for the left panel

        # Draw the left panel
        questionnaire_name = f"{self.full_name} ({self.name})" if self.full_name != self.name else self.name
        num_items = self.num_of_questions
        factors = self.factors.keys()
        self.draw_left_panel(c, width, height, panel_width, questionnaire_name, num_items, factors)

        # Draw the top-left silhouette section
        avg_silhouette, std_dev, num_negative_silhouette = self.calc_silhouette(run_path)
        self.draw_silhouette_section(c, panel_width, width, height, avg_silhouette, std_dev, num_negative_silhouette)

        # Draw the bottom-left alpha section
        cronbach_alpha_total, cronbach_alpha_factors = self.calc_alpha(run_path)
        self.draw_alpha_section(c, panel_width, width, height, cronbach_alpha_total, cronbach_alpha_factors)

        # Draw the top-right domain adaptation section
        self.draw_domain_adaptation_section(c, panel_width, width, height)

        # Calculate model information
        num_models, avg_parameters, std_parameters = self.calc_model_info(run_path)
        # Draw the bottom-right model information section
        self.draw_model_section(c, panel_width, width, height, num_models, avg_parameters, std_parameters)

        # Save the PDF
        c.save()

    def draw_wrapped_text(self, c, x, y, width, text):
        """Utility function to draw wrapped text with increased line spacing."""
        styles = getSampleStyleSheet()
        style = styles["BodyText"]
        style.fontSize = 14  # Increased font size for content text
        style.leading = 18  # Line spacing (increased for better readability)
        para = Paragraph(text, style)
        _, h = para.wrap(width, 200)  # Wrap text within the specified width
        para.drawOn(c, x, y - h)  # Adjust y to prevent overlap
        return h  # Return the height of the wrapped text


    def draw_left_panel(self, c, width, height, panel_width, questionnaire_name, num_items, factors):
        """Draws the left panel with proper spacing for wrapped text."""
        margin = 20
        y_position = height - margin  # Start from the top

        # Draw panel background
        c.setFillColor(colors.lightgrey)
        c.rect(0, 0, panel_width, height, fill=True)

        # Add panel title
        c.setFillColor(colors.black)
        c.setFont("Helvetica-Bold", 18)  # Increased header font size
        c.drawCentredString(panel_width / 2, height - margin - 20, "Questionnaire Summary")
        y_position -= 50

        # Add questionnaire name
        y_position -= self.draw_wrapped_text(c, margin, y_position, panel_width - 2 * margin, f"Name: {questionnaire_name}") + 15

        # Add number of items
        y_position -= self.draw_wrapped_text(
            c,
            margin,
            y_position,
            panel_width - 2 * margin,
            f"Number of Items: {num_items:.4f}" if isinstance(num_items, float) else f"Number of Items: {num_items}",
        ) + 15

        # Add factors
        factors_text = "Factors:<br/>" + "<br/>".join(f"- {factor}" for factor in factors)
        y_position -= self.draw_wrapped_text(c, margin, y_position, panel_width - 2 * margin, factors_text)

        # Draw panel border
        c.setStrokeColor(colors.black)
        c.setLineWidth(1)
        c.rect(0, 0, panel_width, height, fill=False)

    def draw_silhouette_section(self, c, panel_width, width, height, avg_silhouette, std_dev, num_negative_silhouette):
        """Draws the silhouette section with proper spacing."""
        section_width = (width - panel_width) / 2
        section_height = height / 2
        margin = 20
        y_position = height - margin

        # Add section title
        c.setFillColor(colors.black)
        c.setFont("Helvetica-Bold", 18)  # Increased header font size
        c.drawCentredString(panel_width + section_width / 2, height - margin - 20, "Silhouette Information")
        y_position -= 50

        # Add silhouette details
        silhouette_text = (
            f"Average Silhouette: {avg_silhouette:.4f}<br/>"
            f"Standard Deviation: {std_dev:.4f}<br/>"
            f"Number of Questions with Negative Silhouette(Avg per model): {num_negative_silhouette:.4f}"
        )
        self.draw_wrapped_text(c, panel_width + margin, y_position, section_width - 2 * margin, silhouette_text)

        # Draw section border
        c.setStrokeColor(colors.black)
        c.setLineWidth(1)
        c.rect(panel_width, height - section_height, section_width, section_height, fill=False)

    def draw_alpha_section(self, c, panel_width, width, height, cronbach_alpha_total, cronbach_alpha_factors):
        """Draws the Cronbach's Alpha Information section with proper spacing."""
        section_width = (width - panel_width) / 2
        section_height = height / 2
        margin = 20
        y_position = section_height - margin

        # Add section title
        c.setFillColor(colors.black)
        c.setFont("Helvetica-Bold", 18)  # Increased header font size
        c.drawCentredString(panel_width + section_width / 2, section_height - margin - 20, "Cronbach's Alpha Information")
        y_position -= 50

        # Add Cronbach's alpha details
        alpha_text = (
            f"Total Alpha: {cronbach_alpha_total:.4f}<br/><br/>"
            + "Factors:<br/>"
            + "<br/>".join(f"- {factor}: {alpha:.4f}" for factor, alpha in cronbach_alpha_factors.items())
        )
        self.draw_wrapped_text(c, panel_width + margin, y_position, section_width - 2 * margin, alpha_text)

        # Draw section border
        c.setStrokeColor(colors.black)
        c.setLineWidth(1)
        c.rect(panel_width, 0, section_width, section_height, fill=False)

    def draw_model_section(self, c, panel_width, width, height, num_models, avg_parameters, std_parameters):
        """Draws the model information section."""
        section_width = (width - panel_width) / 2
        section_height = height / 2
        margin = 20
        y_position = section_height - margin

        # Add section title
        c.setFillColor(colors.black)
        c.setFont("Helvetica-Bold", 18)  # Increased header font size
        c.drawCentredString(panel_width + section_width + section_width / 2, section_height - margin - 20, "Model Information")
        y_position -= 50

        # Add model statistics
        model_text = (
            f"Number of Models: {num_models}<br/>"
            f"Average Parameters per Model: {avg_parameters:.4f}<br/>"
            f"Standard Deviation of Parameters: {std_parameters:.4f}"
        )
        self.draw_wrapped_text(c, panel_width + section_width + margin, y_position, section_width - 2 * margin, model_text)

        # Draw section border
        c.setStrokeColor(colors.black)
        c.setLineWidth(1)
        c.rect(panel_width + section_width, 0, section_width, section_height, fill=False)

    def draw_domain_adaptation_section(self, c, panel_width, width, height):
        """
        Draws the domain adaptation section in the top-right of the dashboard.
        """
        section_width = (width - panel_width) / 2
        section_height = height / 2
        margin = 20
        y_position = height - margin

        # Add section title
        c.setFillColor(colors.black)
        c.setFont("Helvetica-Bold", 18)  # Increased header font size
        c.drawCentredString(panel_width + section_width + section_width / 2, height - margin - 20, "Domain Adaptation Information")
        y_position -= 50

        # Placeholder values are currently shown.
        domain_text = (
            "Learning Rate: [To be implemented]<br/>"
            "Number of Epochs: [To be implemented]<br/>"
            "Average Change in Mean Score per Epoch: [To be implemented]<br/>"
            "Standard Deviation of Change in Mean Score per Epoch: [To be implemented]"
        )
        self.draw_wrapped_text(c, panel_width + section_width + margin, y_position, section_width - 2 * margin, domain_text)

        # Draw section border
        c.setStrokeColor(colors.black)
        c.setLineWidth(1)
        c.rect(panel_width + section_width, height - section_height, section_width, section_height, fill=False)

    


##################
#   FUNCTIONS   #
##################


def apply_filters(Q, filters_dict):
    return {key: func(Q) for key, func in filters_dict.items()}

def print_permutations(q):
    W = q._pdf['W']
    print(q._descriptor)
    for i, (kmap, w) in enumerate(zip(q._pdf.drop(columns=['P', 'W']).to_dict(orient='records'), W)):
        context = q._context_template.format_map(kmap)
        answer = q._answer_template.format_map(kmap)
        print(f'{i}.',context ,'->', answer, w)

def split_question(Q, index, scales, softmax, filters):
  result = []
  for s in scales:
    q = QCACHE(Q(index=index, scale=s))
    for sf in softmax:
      for f in filters:
        if sf:            
            qsf = QSOFTMAX(q,dim=[index[0], s])
            qsf_f = QFILTER(qsf,filters[f],filtername=f)
            print((index, s),sf,f)
            result.append(qsf_f)
        else:
            qsf = QPASS(q,descupdate={'softmax':''})
            qsf_f = QFILTER(qsf,filters[f],filtername=f)
            print(s,sf,f)
            result.append(qsf_f)
  return result
    
class QMNLIQuestionnaire(Questionnaire):
    
    def run(self, models:List[str]=None,
            questions_ordinals:List[int]=None, 
            result_path=Path('./results/result.csv'), 
            softmax:List[str]=[], 
            filters:Dict[str,dict]={}):
                
        super().run(models = models,
            pipeline_builder = lambda model: pipeline("zero-shot-classification",device=device, model=model),
            questions_ordinals = questions_ordinals,
            result_path = result_path,
            softmax = softmax,
            filters = filters)
        
        
    def question_attributes(self, q):
        attributes = super().question_attributes(q)
        attributes.update({"translation_method" : 'NLI'})        
        return attributes
        
        
class QMLMQuestionnaire(Questionnaire):
    
    def run(self, models:List[str]=None,
            questions_ordinals:List[int]=None, 
            result_path=Path('./results/result.csv'), 
            softmax:List[str]=[], 
            filters:Dict[str,dict]={}):
        
        super().run(models = models,
            pipeline_builder = lambda model: pipeline('fill-mask',device=device, model=model),
            questions_ordinals = questions_ordinals,
            result_path = result_path,
            softmax = softmax,
            filters = filters)
    
    
    def question_attributes(self, q):
        attributes = super().question_attributes(q)
        attributes.update({"translation_method" : 'MLM'})        
        return attributes


from qlatent.qnsp.qnsp import NextSentencePredictionPipeline
class QNSPQuestionnaire(Questionnaire):
    
    def run(self, models:List[str]=None,
            questions_ordinals:List[int]=None, 
            result_path=Path('./results/result.csv'), 
            softmax:List[str]=[], 
            filters:Dict[str,dict]={}):
        
        super().run(models = models,
            pipeline_builder = lambda model: NextSentencePredictionPipeline(model_name=model, device=device),
            questions_ordinals = questions_ordinals,
            result_path = result_path,
            softmax = softmax,
            filters = filters)
        
    
    def question_attributes(self, q):
        attributes = super().question_attributes(q)
        attributes.update({"translation_method" : 'NSP'})        
        return attributes
    

class QCOLAQuestionnaire(Questionnaire):
    
    def run(self, models:List[str]=None,
            questions_ordinals:List[int]=None, 
            result_path=Path('./results/result.csv'), 
            softmax:List[str]=[], 
            filters:Dict[str,dict]={}):
        
        super().run(models = models,
            pipeline_builder = lambda model: pipeline("text-classification",device=device, model=model),
            questions_ordinals = questions_ordinals,
            result_path = result_path,
            softmax = softmax,
            filters = filters)
        
        
    def question_attributes(self, q):
        attributes = super().question_attributes(q)
        attributes.update({"translation_method" : 'COLA'})        
        return attributes
        
##################
# VISUALIZATION  #
##################


def run_models_on_question_and_extract_mean_scores(q, models):
    mean_scores_by_checkpoints={}
    for model in models:
        mean_score=q.run(model).mean_score()
        mean_scores_by_checkpoints[model.checkpoint] = mean_score
    return mean_scores_by_checkpoints

def visualize_checkpoint_scores(q, models):
    checkpoint_scores = run_models_on_question_and_extract_mean_scores(q, models)
    
    constant=checkpoint_scores[0]
    # Sort the dictionary by checkpoint (keys)
    checkpoints = sorted(checkpoint_scores.keys())
    mean_scores = [checkpoint_scores[cp] for cp in checkpoints]
    
    # Calculate y-axis padding based on the range of mean_scores and constant
    all_scores = mean_scores + [constant]
    y_min, y_max = min(all_scores), max(all_scores)
    y_range = y_max - y_min
    
    # Add padding (10% of the range) to y_min and y_max to stretch the plot
    padding = 0.1 * y_range if y_range != 0 else 0.1  # If range is 0, use a small constant padding
    y_min_padded = y_min - padding
    y_max_padded = y_max + padding
    
    # Create the plot
    plt.figure(figsize=(10, 6))
    
    # Plot the checkpoint scores (green line connecting points)
    plt.plot(checkpoints, mean_scores, 'o-', color='green')
    
    # Plot the constant line (red dashed)
    plt.axhline(y=constant, color='red', linewidth=0.8, linestyle='--', label=f'Unbiased mean score (Score={constant})')
    
    # Set x-axis and y-axis ticks to match the dictionary keys and values
    plt.xticks(checkpoints)  # Exact checkpoint values on x-axis
    
    # Make sure the x-axis starts from 0 and is aligned with the y-axis
    plt.xlim(left=0)
    
    # Apply the padded y-axis limits
    plt.ylim(y_min_padded, y_max_padded)
    
    # Add titles and labels
    plt.title('Mean Scores vs Checkpoints')
    plt.xlabel('Checkpoint')
    plt.ylabel('Mean Score')
    
    # Add legend with dynamic labels
    handles, labels = plt.gca().get_legend_handles_labels()
    for cp, score in zip(checkpoints, mean_scores):
        handles.append(plt.Line2D([0], [0], marker='o', color='green', markersize=10, linestyle='-', label=f'Checkpoint {cp} (Score: {checkpoint_scores[cp]})'))
    
    # Display the plot
    plt.grid(True)
    plt.show()



def model_accuracy_on_dataset_per_da_epoch(accuracies : List[float], dataset_name : str) -> None:

    epochs_num = len(accuracies) - 1
    
    title = f"""Accuracy Scores for {dataset_name} in Every DA Epoch"""

    acc_color = "purple"
    
    legend_elements = [
        Patch(facecolor=acc_color, edgecolor=acc_color, label='Accuracy'),
    ]

    fig, ax = plt.subplots()

    ax.legend(handles=legend_elements, loc='center left', bbox_to_anchor=(1, 0.5))

    plt.subplots_adjust(right=0.75)
    
    for epoch_num in range(epochs_num):
        
        acc_xs, acc_ys = [epoch_num, epoch_num + 1], [accuracies[epoch_num], accuracies[epoch_num + 1]]

        plt.plot(acc_xs, acc_ys, color=acc_color)

    all_xs = [i for i in range(epochs_num + 1)]
    plt.scatter(all_xs, accuracies, color=acc_color)
    
    for epoch_num in range(0, epochs_num + 1): 
        plt.plot([epoch_num] * 2, [-0.15, 100.15], color="grey", linestyle="dashed")
    
    plt.plot([-0.5, epochs_num + 0.5], [0] * 2, color="black")
    plt.title(title, loc="left")
    plt.xlabel("DA Epoch Number")
    plt.ylabel("Model Accuracy when Tested on the Mentioned Dataset (%)")
    
    plt.show()
    
    
    

        


import torch
import pandas as pd
import numpy as np
import time
import sklearn as sk
import itertools
import sys
from IPython.core.interactiveshell import InteractiveShell
InteractiveShell.ast_node_interactivity = "all"
from functools import partial
from overrides import override
from typing import *
import copy
from typeguard import check_type
from numbers import Number
from transformers import pipeline
import os
from transformers.tokenization_utils import TruncationStrategy

from ..qabstract.qabstract import *
from ..qabstract.qabstract import SCALE, DIMENSIONS, FILTER, IDXSELECT, _filter_data_frame

## Wait for colab to upgrade to Python 3.11
##IDXSELECT_for_consistency_checks = Annotated[Tuple[Union[slice,Annotated[List[int], MinLen(2)]], MinLen(2)]


class QMNLI(QABSTRACT):

    
    # Static property QMNLI._qregister is defined after this class
    # It will contain one instance (the last one) of every subtype of QMNLI.
    # This will allow to automatically register questions and reuse them with various models when the time comes.

    def __init__(self,
                 context_template: str = " ",
                 answer_template: str = " ",
                 dimensions:DIMENSIONS = {},
                 model: pipeline = None,
                 p=None,
                 index=None,
                 scale='intensifier',
                 descriptor = {}):
        """
        Templates (both context and answer) are parsed before execution to prepare the list of template fields used in each string.
        During execution a list of strings and a list of answer strings are created basaed on cartesian product of the respective field values.
        The field values are extracted from dimensions and intensifiers.
        """        
        super().__init__(dimensions, model, p, index, scale, descriptor)
        
        self._descriptor['query'] = context_template+"->"+answer_template

        self._context_template = context_template
        self._answer_template = answer_template
        QMNLI._qregister[self.__class__.__name__]=self
        

    def run(self, model=None):
        super().run(model)
        if self.model.entailment_id == -1:
            raise Exception("""The entailment id of the MNLI model is not determine.  please update label name to {"CONTRADICTION", "ENTAILMENT", "NEUTRAL"} in self.model.config""")
        T = time.time()
        coo = []
        p = []      

        ## prepare all premise hypothesis pairs
        sequences = []
        for kmap,kcoo in zip(self._keywords_map,self._keywords_grid_idx):
            context = self._context_template.format_map(kmap)
            answer = self._answer_template.format_map(kmap)
            sequences.append((context, answer))
            coo.append(kcoo)
        
        ## Calculate batch entailment probabilities
        inputs = self.model.tokenizer(
                sequences,
                add_special_tokens=True,
                return_tensors="pt",
                padding=True,
                truncation=TruncationStrategy.ONLY_FIRST,
            )
        
        model_inputs = {k: inputs[k].to(self.model.device) for k in self.model.tokenizer.model_input_names}
        results = self.model.model(**model_inputs)
        
        ## Apply softmax on logits
        for res, (context, answer) in zip(results.logits, sequences):  
            output = {
                'logits': res.detach().cpu().view(1, -1),
                "candidate_label": answer,
                "sequence": context,
            }
            p.append(self.model.postprocess([output], multi_label=False)['scores'][0])

        
        coo = torch.stack(coo).T
        assert torch.all(torch.eq(coo.T, self._keywords_grid_idx))
        
        self._pdf["P"] = p
        
        self._T = time.time() - T
        self.result = self
        return self.result

    
# The static property QMNLI._qregister will contain one instance (the last one) of every subtype of QMNLI.
# This will allow to automatically registering questions and reusing them with various models when the time comes.


QMNLI._qregister:Dict[str,QMNLI]={}


class _QMNLI(QMNLI):
    
    index = ["emotion"]
    scale = "intensifier"

    def __init__(self, context, template, emo_pos, emo_neg, intensifiers, **kwargs):
        
        super().__init__(
            context_template=context,
            answer_template=template,
            dimensions={"emotion":dict_pos_neg(emo_pos, emo_neg,1.0), "intensifier":intensifiers, },
            **kwargs
        )

        # self.scale = "intensifier"
        # self.index = "emotion"
        self.emo_pos = emo_pos
        self.emo_neg = emo_neg
        self.grouping = [{"emotion":emo_pos},{"emotion":emo_neg}]
        

    def report(self,scale:Union[str,int]="intensifier", index="emotion", filters:Dict[str,FILTER]={"unfiltered":{}}, grouping:List[FILTER]=[],):
        if len(grouping)==0:
            grouping = self.grouping
        return super().report(scale=scale,index=index,filters=filters,grouping=grouping)


    def softmax(self, dim:Union[str,int,List[str],List[int]]="intensifier"):
        return super().softmax(dim=dim)

    
    def to_dataframe(self, scale:Union[str,int]="intensifier", index:List[str]="emotion", filter:FILTER={}, categories:Dict[str,Dict[str,str]]={}):
        return super().to_dataframe(scale=scale, index=index, filter=filter, categories=categories )



import torch
import pandas as pd
import numpy as np
import time
from pprint import pprint
from transformers import AutoModelForMaskedLM, AutoTokenizer
from transformers import pipeline
from transformers import PreTrainedModel
from transformers import PreTrainedTokenizer
import scipy
import sklearn as sk
import itertools

from ..qabstract.qabstract import *
from ..qabstract.qabstract import SCALE, DIMENSIONS, FILTER, IDXSELECT, _filter_data_frame


class QCOLA(QABSTRACT):

    # Static property QCOLA._qregister is defined after this class
    # It will contain one instance (the last one) of every subtype of QCOLA.
    # This will allow to automatically register questions and reuse them with various models when the time comes.

    def __init__(self,
                 template: str,
                 dimensions:DIMENSIONS = {},
                 model: pipeline = None,
                 p=None,
                 index=None,
                 scale='intensifier',
                 descriptor = {}):
        super().__init__(dimensions, model, p, index, scale, descriptor=descriptor)
        
        self._index = index
        self._scale = scale
        self._descriptor['query'] = template
        self._template = template
#         self.intensifier_names = list(dimensions[scale].keys())
#         self.emotions = list(dimensions[index].keys())
        QCOLA._qregister[self.__class__.__name__] = self

        
    def run(self, model= None):
        super().run(model)
        T = time.time()
        coo = []
        p = []
        for kmap,kcoo in zip(self._keywords_map,self._keywords_grid_idx):
            query = self._template.format_map(kmap)
            ans = self.model(query)[0].get('score')
            p.append(torch.Tensor([ans]).squeeze().cpu().item())
            coo.append(kcoo)

        coo = torch.stack(coo).T
        assert torch.all(torch.eq(coo.T, self._keywords_grid_idx))

        self._pdf["P"] = p
        
        self._T = time.time() - T
        self.result = self
        return self.result


    def ans_logits(self, result):
        """
        result = [{'label': 'LABEL_1', 'score': 0.8850699663162231}, ... ]
        """
        ans = [r['score'] if r['label'] == 'LABEL_1' else 1 - r['score'] for r in result]
        return ans


from typing import Dict

# The static property QCOLA._qregister will contain one instance (the last one) of every subtype of QCOLA.
# This will allow to automatically registering questions and reusing them with various models when the time comes.
QCOLA._qregister: Dict[str, QCOLA] = {}




from setuptools import setup, find_packages
from pathlib import Path
from pkg_resources import parse_requirements

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8-sig")
setup(
    name="qlatent",
    version="1.0.18",
    description="A Python package for running psychometric on LLMs.",
    packages=find_packages(),
    classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: OS Independent",
    ],
    include_package_data=True,
    python_requires=">=3.8",
    install_requires=[
    "numpy==1.24.4",
    "pandas",
    "seaborn",
    "scipy",
    "scikit-learn",
    "transformers",
    "pingouin",
    "typeguard",
    "datasets",
    "evaluate",
    "matplotlib",
    "sentence_transformers",
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    data_files=[('Academic Article', ['2409.19655v1.pdf'])],
)


import numpy as np
import pandas as pd
import seaborn as sns
import torch
import scipy
import time
import sklearn as sk
from pprint import pprint
from transformers import AutoModelForMaskedLM, AutoTokenizer
from transformers import pipeline
from transformers import PreTrainedModel
from transformers import PreTrainedTokenizer
from abc import ABC, abstractmethod
import copy
from overrides import overrides
from typing import * 
from numbers import Number
import itertools
from typeguard import check_type
from functools import partial
import pingouin as pg


SCALE = Dict[str,Number]
DIMENSIONS = Dict[str,SCALE]
FILTER = Dict[str,Collection[str]]
IDXSELECT = Tuple[Union[slice,List[int]]]


def dict_same_weight(w,ks):
  return dict(zip(ks,[w]*len(ks)))


def dict_pos_neg(pos, neg, w):
  return dict(dict_same_weight(1.0*w/len(pos),pos), **dict_same_weight(-1.0*w/len(neg),neg))


def _filter_tensor(t, slices):
    for i in range(len(slices)):
        t = t[slices[i]]
    return t


def _filter_data_frame(df, filter:FILTER):
    select = df.copy()
    for f in filter: #only select rows where df[f] in filter[f]
        select[f] = select[f].apply(lambda x: x in filter[f])
    select = select.all(axis=1)
    return select
    

def fixed_check_type(var, expected_type):
    try:
        check_type(var, expected_type)
        return True
    except:
        return False


# def check_type(var, expected_type):
#   return isinstance(var, expected_type)


def print_gradient(df):
    import seaborn as snsintensifier_names
    cm = sns.light_palette("green", as_cmap=True)
    s = df.style.background_gradient(cmap=cm, axis=None) 
    s = s.format(precision = 4)
#     s = s.set_precision(4)
    return s


def wrap_replace_callable(c, f, with_copy=False, *fargs, **fkwargs):
    def wrapper(*cargs, **ckwargs):
        rc = c(*cargs, **ckwargs)
        rf = f(rc, *fargs, **fkwargs)
        return rf
    if with_copy:
        q = copy.deepcopy(c)
    else:
        q = c
    q.__call__ = wrapper
    return q
        

class QABSTRACT(ABC):

    
    def __init__(self, dimensions:DIMENSIONS = {}, model: pipeline = None, p=None, index=None, scale='intensifier', descriptor = {}):
        
        self._dimensions = dimensions
        self._field_names = list(self._dimensions.keys())
        self._scale = scale
        self.result = None
        self._index = index if index is not None else list(set(self._field_names) - set([scale]))
        self._descriptor = descriptor
        self._descriptor['scale'] = str(self._scale)
        self._descriptor['index'] = str(self._index)

        self._keywords = {d:list(scale.keys()) for d,scale in dimensions.items()}
        #print(f"_keywords={self._keywords}")
        self._keywords_indices = {d:dict([(k,i) for i,k in enumerate(self._keywords[d])]) for d in self._keywords}
        #print(f"_keywords_indices={self._keywords_indices}")
        self._dimshape = tuple([len(self._keywords[d]) for d in self._field_names])
        #print(f"_dimshape={self._dimshape}")

        self._keywords_grid = list(itertools.product(*[self._keywords[f] for f in self._field_names]))
        #print(f"_keywords_grid={self._keywords_grid}")
        self._keywords_map = [dict(zip(self._field_names,k)) for k in self._keywords_grid]
        #print(f"_keywords_map={self._keywords_map}")
        self._keywords_grid_idx = torch.Tensor([tuple([self._keywords_indices[f][ktuple[i]] for i,f in enumerate(self._field_names)]) for ktuple in self._keywords_grid])
        #print(f"_keywords_grid_idx={self._keywords_grid_idx[:3]}")


        W = [tuple([self._dimensions[f][ktuple[i]] for i,f in enumerate(self._field_names)]) for ktuple in self._keywords_grid]
        #print(f"W={W}")
        self._weights_grid = pd.DataFrame(W, columns=self._field_names)
        #print(f"_weights_grid={self._weights_grid}")
        self._weights_flat = self._weights_grid.prod(axis=1)
        #print(f"_weights_flat={self._weights_flat}")
        self._weights_dense = torch.sparse_coo_tensor(self._keywords_grid_idx.T, self._weights_flat, self._dimshape).to_dense()
        #print(f"_weights={self._weights.shape}")


        self._pdf = pd.DataFrame(self._keywords_grid, columns=self._field_names)
        self._pdf = self._pdf.assign(P=0)
        self._pdf = self._pdf.assign(W=self._weights_flat)
        #print(f"_pdf={self._pdf}")
        

        self._p = p
        self.model = model
        

    @abstractmethod
    def run(self, model):
        if model:
            self.model = model
        


    def _grouping_suitable_for_consistency_check(self, grouping, filter={}):
        if len(grouping)<2:
            return False
        
        label_count = 0
        
        for group_filter in grouping:
            df = self._pdf[_filter_data_frame(self._pdf, filter)]
            df = df[_filter_data_frame(df,group_filter)]
            if len(df) > 0:
                label_count += 1
                
        return label_count >= 2


    def _filter_words_to_slice(self, filter:FILTER)->IDXSELECT:
        s = [ #loop over dimensions
                        slice(None,None) if d not in filter else
                        [ #loop over keywords in dimesion
                            self._keywords_indices[d][k]
                            for k in filter[d]
                        ]
                        for d in self._field_names
                      ]
        #one-hot econding, otherwise the shapes won't broadcast together...
        n = len(self._field_names)
        result = [[slice(None,None,None) for _ in range(n)] for _ in range(n)]
        for i in range(n):
            result[i][i]=s[i]
        return result


    def _create_default_grouping(self, group_field):
        default_grouping_df = pd.DataFrame(self._dimensions[group_field].items(), columns=['name', 'score'])
        grouping = [{group_field: group_df['name'].tolist()} for group_score, group_df in default_grouping_df.groupby('score')]
        # print(f'No grouping stated, using default grouping on {group_field}:')
        # pprint(grouping)
        return grouping
        
    
    def _pd_values_sort_key(self, terms:pd.Series)->pd.Series:
        f = terms.name
        result = terms.apply(lambda x: self._dimensions[f][x])
        return result


    def get_filter_for_postive_keywords(self, ignore_set=None):
        ignore_set = {self._scale} if ignore_set is None else ignore_set
        result:FILTER={}
        for dim in self._keywords:
            result[dim]=[]
            for k in self._keywords[dim]:
                if self._dimensions[dim][k] > 0 or dim in ignore_set:
                    result[dim].append(k)
        return result

    
    def __add__(self,other):
        result = copy.deepcopy(self)
        result._p += other._p
        result._p_raw += other._p_raw
        return result

    
    def __sub__(self,other):
        result = copy.deepcopy(self)
        result._p -= other._p
        result._p_raw -= other._p_raw
        return result

    
    def __mul__(self,other):
        result = copy.deepcopy(self)
        result._p *= other._p
        result._p_raw *= other._p_raw
        return result

    
    def __truediv__(self,other):
        result = copy.deepcopy(self)
        result._p /= other._p
        result._p_raw /= other._p_raw
        return result

    
    def __call__(self, model=None):
        return self.run(model)


    # # def __str__(self):
    # #   return f"<{self.__class__.__name__} object> {self._descriptor}"
    # def __repr__(self):
    #   return f"<{self.__class__.__name__} object> {self._descriptor}"

    
    def __hash__(self):
        """
        Questions having the same are descriptor should represent the same question.
        """
        return hash(frozenset(self._descriptor.items()))
    
    
    def __eq__(self,other):
        """
        Questions having the same are descriptor represent the same question.
        """
        return self._descriptor == other._descriptor

    
    def set_model(self, model):
        self.model = model
       
          
    def softmax(self, dim=0, temperature = 1):
        """
        params:
        dim: int or tuple. The dimensions to normalize with softmax. Dimesions will be normalized one by one in the given order.
        """
        if not fixed_check_type(dim, List):
            dim = [dim]
        if fixed_check_type(dim, List[str]):
            dim = [self._field_names.index(d) for d in dim]
        elif fixed_check_type(dim, List[int]):
            pass
        else:
            raise TypeError(dim)

        result = copy.deepcopy(self,memo={id(self.model):self.model})
        # del result.model
        # result.model = self.model
        coo = self._keywords_grid_idx.T
        p = torch.Tensor(self._pdf["P"])
        p_dense = torch.sparse_coo_tensor(coo, p, self._dimshape).to_dense()

        p_sparse = p_dense.to_sparse()
        p_vals = p_sparse.values()
        p_coos = p_sparse.indices()
        assert torch.all(torch.eq(p, p_vals))
        assert torch.all(torch.eq(coo, p_coos))
                
        for d in dim:
            p_log = torch.log(p_dense)
            p_log = p_log / temperature
            p_dense = torch.nn.functional.softmax(p_log, dim=d)
            

        p_vals = p_dense.to_sparse().values()

        result._pdf["P"] = p_vals
        return result    


    def minmax(self, dim=Union[str,int,List[str],List[int]]):
        """
        params:
        dim: int or tuple. The dimensions to normalize with softmax. Dimesions will be normalized one by one in the given order.
        """
        if not fixed_check_type(dim, List):
            dim = [dim]
        if fixed_check_type(dim, List[str]):
            dim = [self._field_names.index(d) for d in dim]
        elif fixed_check_type(dim, List[int]):
            pass
        else:
            raise TypeError(dim)

        result = copy.deepcopy(self,memo={id(self.model):self.model})
        coo = self._keywords_grid_idx.T
        p = torch.Tensor(self._pdf["P"])
        p_dense = torch.sparse_coo_tensor(coo, p, self._dimshape).to_dense()

        p_sparse = p_dense.to_sparse()
        p_vals = p_sparse.values()
        p_coos = p_sparse.indices()
        assert torch.all(torch.eq(p, p_vals))
        assert torch.all(torch.eq(coo, p_coos))

        print(p_dense.shape)
        for d in dim:
            print(d)
#             p_dense = torch.nn.functional.softmax(p_dense, dim=d)
            p_dense = p_dense / torch.max(p_dense, dim=d).to_sparse().repeat(*list(set(dim) - set([d])))
            print(p_dense.shape)

        p_vals = p_dense.to_sparse().values()

        result._pdf["P"] = torch.Tensor(p_vals)
        return result
        
    
    def to_dataframe(self, scale:Union[str,int], index:List[str], filter:FILTER={}, categories:Dict[str,Dict[str,str]]={}):
        # try:
        #     check_type(scale, int)
        #     scale = self._field_names[scale]
        # except TypeCheckError:
        #     check_type(scale, str) #redundant check for clarity
        
        if fixed_check_type(scale, int):
          scale = self._field_names[scale]
        elif fixed_check_type(scale, str):
            pass
           #redundant check for clarity
        else:
            raise TypeError(scale)


        select = _filter_data_frame(self._pdf, filter)
        df = self._pdf[select]
        print(f"index = {index}")
        print(set(self._field_names), set([scale]), set(index) )
        aggregated_fields = set(self._field_names) - set([scale]) - set(index)
        aggregated_fields = list(aggregated_fields)
        print(aggregated_fields)
        if len(aggregated_fields)>0:
            W = self._weights_grid[select]
            W = W[aggregated_fields]
            W = W.prod(axis=1)
            df["P"] = df["P"]*W


        for f in categories.items(): #replace individual keywords with categories
            df[f] = df[f].apply(lambda x: categories[f][x])

        #group by index categories
        df = pd.pivot_table(df, values='P', index=index, columns=[scale], aggfunc=np.mean)
        df = df.sort_index(axis=1, key=self._pd_index_sort_key)
        df = df.sort_index(axis=0, key=self._pd_index_sort_key)

        return df

        
    def _pd_index_sort_key(self, terms:pd.Index)->pd.Index:
        f = terms.name
        result = pd.Index([self._dimensions[f][x] for x in terms])
        return result


    def mean_score(self, filter:FILTER={}):
        select = _filter_data_frame(self._pdf,filter)
        P = self._pdf[select]
        score = P["P"]*P["W"]
        score = score.mean()
        return score
    
    
    def effect_size(self, filter:FILTER={}):
        select = _filter_data_frame(self._pdf,filter)
        P = self._pdf[select]
        score = P["P"]*P["W"]
        
        effect_size = score.mean() / score.std()
        return effect_size


    def internal_consistency(self, measure="silhouette_score", metric="correlation", grouping:List[FILTER]=[], scale:Union[str,int] = None, index:List[str] = None, filter:FILTER={}):
        """
        Before the internal consistency check the data is diced according to the filter -- e.g. only entries matching the filter will be considered.
        This operation first splits the data into scale long vectors. The vectors are grouped according to the grouping argument.
        The vectors within each group are expected to be similar (high correlation, low distance) while vectors in different groups are expected to be different.
        This expectation is validated by Scikit's unsupervised cluster quality measures.

        Parameters:
        -----------
        @filter
            Specifies a set of keywords to keep along chosen dimensions. Unspecified dimensions are kept (not filtered out).
        @scale
            Dimension specifying the vector for which distances are computed. This is the dimension which is used for likert scale values.
        @grouping
            Specifies multiple dices (each one is structured like the filter) within which likert scale values are expected to be the same.
            Likert scale values are expected to be different for different dices. Dices are not required to have the same shape.
        @measure: string
            An unsupervised cluster quality measure from scikit. Must be one of 'calinski_harabasz_score', 'davies_bouldin_score', 'silhouette_score', 'silhouette_samples'
        @metric: string
            One of distance metrics from 'sklearn.metrics.pairwise_distances' used for computing silhouette
        """
#         try:
#           scale = self._scale if scale is None else scale
#           check_type(scale, int)
#           scale = self._field_names[scale]
#         except TypeCheckError:
#           check_type(scale, str) #redundant check for clarity

        scale = self._scale if scale is None else scale
        if fixed_check_type(scale, int):
            scale = self._field_names[scale]
        elif fixed_check_type(scale, str):
            pass
           #redundant check for clarity
#         else:
#             raise TypeError(scale)

        # Add default grouping for the index field
        index = self._index if index is None else index
        if grouping is not None and len(grouping) == 0:
            group_field = index[0]
            grouping = self._create_default_grouping(group_field)

        df = self._pdf[_filter_data_frame(self._pdf, filter)]

        X = []
        L = []
        for label,group_filter in enumerate(grouping):
            dice = df[_filter_data_frame(df, group_filter)]
            dice = dice.pivot_table(index=index,columns=scale, values="P", aggfunc='mean')
            X += [dice]
            L += [label]*len(dice)
        X = pd.concat(X, axis=0)

        f = getattr(sk.metrics, measure)
        if 'metric' in f.__code__.co_varnames[:f.__code__.co_argcount + f.__code__.co_kwonlyargcount]:
            result = f(X=X, labels=L, metric=metric)
        else:
            result = f(X=X, labels=L)

        return result

    
    def print_gradient(self, raw_probabilities=False):
        df = self.to_dataframe(self, raw_probabilities=raw_probabilities)
        cm = sns.light_palette("green", as_cmap=True)
        return df.style.background_gradient(cmap=cm, axis=None)

    
    def inner_alpha(self, filter:FILTER={}):
        df = self.to_dataframe(scale=self._scale, index=self._index, filter=filter)
        if hasattr(self, '_filter') and bool(self._filter) and self._index[0] in self._filter:
            indecies_list = self._filter[self._index[0]]
            dict_items = self._dimensions[self._index[0]].items()
            matches = [(key, value) for key, value in dict_items if key in indecies_list]
            default_grouping_df = pd.DataFrame(matches, columns=['name', 'score'])
        else:
            default_grouping_df = pd.DataFrame(self._dimensions[self._index[0]].items(), columns=['name', 'score'])
        grouping = [{self._index[0]: group_df['name'].tolist()} for group_score, group_df in default_grouping_df.groupby('score')]
        group_scores = pd.DataFrame(self._dimensions[self._index[0]].items(), columns=['name', 'score'])['score']
        for group in grouping:
            vals = group[self._index[0]]
            print(vals, 'Alpha:',pg.cronbach_alpha(data=df.T[vals])[0])
        if not (hasattr(self, '_filter') and bool(self._filter)):
            a3 = pg.cronbach_alpha(pd.DataFrame((df.T.to_numpy() * group_scores.to_numpy())))
            print('Global alpha:', a3[0])
            return a3[0]
        return pg.cronbach_alpha(data=df.T[vals])[0]
        

    def report(self,scale:Union[str,int]=None, index:List[str]=None, filters:Dict[str,FILTER]={"unfiltered":{}}, grouping:List[FILTER]=[],):
        scale = self._scale if scale is None else scale
        if not fixed_check_type(scale, str):
            scale = self._field_names.index(scale)
        if index is None:
            if self._index:
                index = self._index
            else:
                index = list(set(self._field_names) - set([scale, "P", "W"]))

        print(f"Query time: {self._T}")
        for label, filter_dict in filters.items():
            print(f"Mean score {label} [{self._weights_flat.min()}..{self._weights_flat.max()}]: {self.mean_score(filter=filter_dict)}")

         # Add default grouping for the index field
        if grouping is not None and len(grouping) == 0:
            group_field = index[0]
            grouping = self._create_default_grouping(group_field)
        
        for label, filter_dict in filters.items():
            if self._grouping_suitable_for_consistency_check(grouping, filter=filter_dict):
                partial_internal_consistency = partial(self.internal_consistency, grouping=grouping,filter=filter_dict, index=index , scale=scale)
                print(f"Internal consistency (silhouette, correlation) for {label}: {partial_internal_consistency(measure='silhouette_score', metric='correlation')}")
                print(f"Internal consistency (Calinski&Harabasz)  for {label}: {partial_internal_consistency(measure='calinski_harabasz_score')}")
                print(f"Internal consistency (Davies&Bouldin) for {label}: {partial_internal_consistency(measure='davies_bouldin_score')}")
            else:
                print("At least two groups with at least two vectors in each group should be specified to check for internal consistency.")

        for label, filter_dict in filters.items():
            print("\n")
            display(
              print_gradient(
                  self.to_dataframe(
                    scale=scale,
                    index=index,
                    filter=filter_dict,
            )))
            self.inner_alpha(filter=filter_dict)
        


class QDELEGATOR(QABSTRACT):

    
    def __init__(self, srcobj):
        """
        Shallow copy the state of the source object.
        Both objects will share the same current state.
        The run method of QDELEGATOR delegates the execution to srcobj.run and
        copies the state after each execution. Any modifications made to the state
        of srcobj prior to execution of run(..) are copied as well.
        Further modifications to the shallow state of either one of the objects will
        not affect the other one.
        @srcobj
          initialized object of a question.
        """
        super().__init__()
        self.__dict__.update(srcobj.__dict__)
        self._run = srcobj.run
        
    
    @overrides
    def run(self, model=None, **kwargs):
        #print(f"{self.__class__.__name__} delgates execution of run(..) to {self._run.__self__.__class__.__name__}" )
        result = self._run(model, **kwargs)
        d = dict(result.__dict__)
        #retain self._descriptor and self._run
        if "_descriptor" in d: del d['_descriptor']
        if "_run" in d: del d['_run']
        self.__dict__.update(d)
        # #restore self._run to maintain full delegation path
        # self._run = result.run
        return self


class QCACHE(QDELEGATOR):

    
    def __init__(self, srcobj, **kwargs):
        super().__init__(srcobj, **kwargs)
        #subclasses instantiated with questions that were already run
        #may set the cached flag to False to force rerunning the question
        self._cached = False
        self._last_model_identifier = None

    
    @overrides
    def run(self,model=None, **kwargs):
        # print(f"Execute cachable question with model {model}.")
        # print(f"Cached: {self._cached}, same model:{model.model_identifier == self._last_model_identifier}.")
        if self._cached and model.model_identifier == self._last_model_identifier:
            # print(f"Skipping model execution. Use cached results.")
            return self
        result = super().run(model=model,**kwargs)
        self._cached = True
        self._last_model_identifier = model.model_identifier
        assert result==self
        return self



class QPASS(QDELEGATOR):

    
    def __init__(self, qobject, descupdate={}):
        super().__init__(qobject)
        #copy descriptor from srcobj to avoid propagation of changes
        self._descriptor = dict(self._descriptor, **descupdate)


class QSOFTMAX(QPASS):

    
    def __init__(self, qobject, dim="intensifier", temperature=1):
        super().__init__(qobject, {'softmax':str(dim)})
        self._dim = dim
        self._temperature = temperature

    
    @overrides
    def run(self,model=None, **kwargs):
        result = super().run(model, **kwargs)
        return result.softmax(dim=self._dim, temperature=self._temperature)


class QMINMAX(QPASS):

    
    def __init__(self, qobject, dim="intensifier"):
        super().__init__(qobject, {'minmax':str(dim)})
        self._dim = dim

    
    @overrides
    def run(self,model=None, **kwargs):
        result = super().run(model, **kwargs)
        return result.minmax(dim=self._dim)


class QFILTER(QPASS):

    
    def __init__(self, qobject, filter:FILTER = {}, filtername=""):
        super().__init__(qobject, {'filter':filtername})
        self._filter = filter

    
    @overrides
    def run(self,model=None, **kwargs):
        result = super().run(model, **kwargs)
        assert result == self
        select = _filter_data_frame(self._pdf, self._filter)
        result._pdf = self._pdf[select]
        result._weights_grid = self._weights_grid[select]
        return result




######################BuildModelLabels######################
import torch
from transformers import pipeline
import pandas as pd
import numpy as np
from typing import Callable, List, Dict
import os
import warnings
import gc
os.environ['CUDA_LAUNCH_BLOCKING'] = '1'
warnings.filterwarnings('ignore')
from collections import Counter
######################BuildModelLabels######################



class BuildModelLabels:
    def __init__(self, model_name : str,
                 label_2_dataset_id : dict = {"entailment": 0, "neutral" : 1, "contradiction" : 2}):
        
        self.model_name=model_name.replace("/","_",1)
        self.label_2_dataset_id = label_2_dataset_id
        self.data_set_path = os.path.join(os.path.dirname(__file__), 'mnli_label_detection_dataset')
        self._build_predictions_dict()
    
    def _build_predictions_dict(self):
        self.predictions_dict = {key:[] for key in self.label_2_dataset_id}
        
    def _get_names(self, directory_path : str, ending : str) -> List[str]:
        """
            Return a list of the names of all files with a specific ending that are inside a given directory.
        """

        names_list = []
    
        for filename in os.listdir(directory_path):
            if filename.endswith(ending):
                names_list.append(filename[:-(len(ending)+1)]) # Dont include ending

        return names_list
    
    
    def _get_split_length(self,split_name : str) -> int: # works
        """
        Returns the number of rows of the specified split.
        """

        df = pd.read_csv(os.path.join(self.data_set_path, f"{split_name}.csv"), encoding = "utf-8-sig")
        row_count = len(df)
        return row_count
    
    
    def _load_k_rows(self, split_name : str, k : int,total_predictions) -> pd.DataFrame:
        """
        Returns a dataframe that contains k new rows of the split $split_name.
        """

        header_names = ['premise', 'hypothesis', 'genre', 'label']
        k_rows_df = pd.read_csv(os.path.join(self.data_set_path, f"{split_name}.csv"),
                    encoding = "utf-8-sig",
                    header = None,
                    names = header_names,
                    skiprows = 1 + total_predictions,
                    nrows=k)

        return k_rows_df # THE BATCH TO BE CLASSIFED

    
    
    def _predict_k_rows(self, split_name : str, predict_batch : Callable[[List[str]], List[int]], k : int, total_predictions) -> None:
        """
        Predicts the label of $k rows (premise hypothesis pairs)
        And increases the split_index and correct_predictions of the $model csv file.
        """

        rows_df = self._load_k_rows(split_name, k, total_predictions)
        premises, hypotheses, true_labels = [], [], []
        for row in rows_df.itertuples():
            premises.append(row.premise)
            hypotheses.append(row.hypothesis)
            true_labels.append(row.label)

        predicted_labels = predict_batch(premises, hypotheses)    
        self.predictions_dict[split_name]=self.predictions_dict[split_name]+predicted_labels
        
        #correct_predictions = sum([predicted_labels[i] == true_labels[i] for i in range(k)])
        total_predictions += k
        return total_predictions
                    

    
    def _predict_function(self):
        
        # Clear memory
        torch.cuda.empty_cache()
        gc.collect()
        mnli = pipeline("zero-shot-classification", device=torch.device(0 if torch.cuda.is_available() else 1), model=self.model_name.replace('_', '/',1))
        if hasattr(mnli.model.config, 'id2label'):
            print(f"{self.model_name} ORIGINAL CONFIG:\n {mnli.model.config.id2label}")
        else:
            print(f"{self.model_name} original config is unknown.")
        def predict_batch(premises: List[str], hypotheses: List[str]) -> List[int]:
            """
                Uses model given create_predict_function to predict a batch of premise&hypothesis pairs.
            """        
            # Initialize a list to store the predicted labels
            predicted_ids = []
            # Tokenize the batch of premises and hypotheses
            inputs = mnli.tokenizer(premises, hypotheses, truncation=True, max_length=1024, padding=True, return_tensors='pt')
            # Move inputs to CUDA if available
            model_inputs = {k: v.to('cuda') for k, v in inputs.items()}
            # Forward pass through the model
            with torch.no_grad():
                outputs = mnli.model(**model_inputs)
                # Calculate probabilities and predict labels
                probs = torch.softmax(outputs.logits, dim=1).cpu().detach().numpy()
                batch_predicted_ids = np.argmax(probs, axis=1).tolist()
                # Append batch predictions to the list of predicted labels
                predicted_ids.extend(batch_predicted_ids)

            return predicted_ids
        return predict_batch
    
    
    def _perform_predictions(self):
        splits_names = self._get_names(self.data_set_path,'csv')
        predict_function = self._predict_function()
        batch_size = 64
        for split_name in splits_names:
            total_predictions = 0
            split_length = self._get_split_length(split_name)
            k = min(split_length - total_predictions, batch_size) 
            while k > 0:
                total_predictions = self._predict_k_rows(split_name, predict_function, k, total_predictions)
                k = min(split_length - total_predictions, batch_size)
    
    def return_id2label(self):
        self._perform_predictions()
        splits_names = self._get_names(self.data_set_path,'csv')
        id2_label={}
        for split_name in splits_names:
            
            numbers = [int(x) for x in self.predictions_dict[split_name]]
            # Use Counter to count occurrences of each number
            counter = Counter(numbers)

            # Use max() function with key argument to find the most common number
            most_common_number = max(counter, key=counter.get)
            id2_label[most_common_number]=split_name
        print("============NEW MODEL CONFIG===========")
        print(id2_label)
        print("========================================")
        return id2_label


